// SP70: a program for unsupervised learning in the ICMAUS framework.

// Copyright (C) 2003 J G Wolff.

// This program is released to the public domain without any restrictions.
// You can redistribute it and/or modify it as you wish.

// In any publication that refers to any version of this program, please
// acknowledge J G Wolff as the originator of the program. 

// A description of this program may be found in "Unsupervised learning
// in a framework of information compression by multiple alignment,
// unification and search", J G Wolff, March 2002. A copy may be obtained
// from http://uk.arxiv.org/abs/cs.AI/0302015. See also
// http://www.cognitionresearch.org.uk/papers/ul/ul.htm.

// Please retain this notice on all versions.

// Dr J G Wolff, www.CognitionResearch.org.uk.
// Email: jgw@cognitionresearch.org.uk.

// May 2003.

#define PROGRAM "SP70"

#define VERSION "9.2"

/*
**
** ABSTRACT:	Development of the SP model for unsupervised inductive learning.
**
** ENVIRONMENT: The module requires the header file "SP70_head.h" 
**				to be on the `include' path.  It also makes use of 
**				routines defined in "SP70_lib.cpp"
**
** AUTHOR:		Gerry Wolff
**
** MODIFIED BY:
**
** SP70:
**
** 2.0 JGW 7/4/99 - 	Derived from SP61, v 4.5. See sp70_od, %9.
** 2.1 JGW				Backup
** 2.2 JGW 10/8/99 -	See sp70_od, %10 and %11. This version is
**						confined to the creation of new patterns
**						based on alignments of 'raw' data. This
**						is a stepping stone towards the development 
**						of a system that can learn higher-level
**						patterns. This version is mainly a tidying up
**						of the compress function and functions called
**						from there.
** 2.3 JGW 14/9/99 -	See sp70_od, %13, %14 and %15. Put codes 
**						on all alignments with New that are not
**						'recognised' and accommodate sequences of
**						patterns. This version removes the anomaly
**						that original patterns had a depth of 0.
**						They now have a depth of 1.
** 2.4 JGW 1/10/99 -	Introducing unifications as discussed in
**						sp_70_od, %16 and %17.
** 2.5 JGW 10/11/99 -	Introducing distinction between
**						CONTENTS and IDENTIFICATION
**						symbols and the way it constrains matching,
**						as in SP61, v 5.0.
** 2.6 JGW 11/11/99 -	Development of the way New is moved to Old.
** 2.7 JGW 13/11/99 -	Further development of the learning function.
**						See sp70_od, %19.
** 2.8 JGW 18/11/99 -	New patterns not added to Old until a check has
**						been made that there is not already a pattern
**						in Old with the same CONTENTS symbols.
** 3.0 JGW 22/2/00 -	[Borland C++] Described in sp70_od, %22. The scoring system
**						for alignments will be changed so that it is
**						based directly on the actual or calculated size
**						of the code which would be needed to compress
**						new in terms of the proposed alignment.
**						Calculations which use gaps will not be used
**						except, possibly for the interim calculations
**						which are made while the hit structure is
**						being built.
**						In this version, all that has been done in this
**						direction is to collapse the two versions of
**						current_new_pattern to one and correct
**						some bugs.
** 3.1 JGW 25/2/00 -	[Borland C++] Re-jigs the 'learning' of new patterns to focus
**						on the creation of 'unified' patterns as
**						described in sp70_od, %24. As an experimental
**						step, *all* alignments, other than those which
**						are the result of unifications, are excluded
**						from the alignment process (see the note at
**						the end of sp70_od, %24).
** 3.2 JGW 8/5/00 - 	[Borland C++] Every time a New pattern is added to Old it is given
**						code symbols and a corresponding value for its
**						coding cost. Alignments are formed for unifications
**						of patterns as described in sp70_od, %28.
** 3.3 JGW 16/5/00 -	[Borland C++] Patterns that are newly-formed as a result of
**						unifications are matched amongst themselves with
**						the aim of forming further unifications and also
**						with the aim of minimising the number
**						of IDENTIFICATION symbol
**						types that need to be used (see sp70_od, %32, %33, %34).
**						The earlier version of compute_score_with_gaps(), which takes
**						account of gaps, has been re-instated
**						(see sp70_od, %35) and then revised in the light of
**						the distinction between IDENTIFICATION and
**						CONTENTS symbols.
** 3.4 JGW 26/5/00 -	[Borland C++] Further development as described in sp70_od, %41.
**						During learning, the program checks for 'wild' matches
**						between patterns before code symbols are assigned. This
**						helps to minimize the number of different code symbol
**						numbers that are used. The program also checks for
**						exact matches after new code symbols are assigned.
** 3.5 JGW 7/6/00 - 	[Borland C++] This version makes the non-unified alternatives in
**						an alignment into the 'referents', referred to from
**						the unified pattern (sp70_od, %45). Discrimination
**						symbols are used to distinguish the alternatives in
**						a given context. This version is left at a stage
**						where it can encode unified and non-unified patterns
**						but has not sorted out discrimination symbols or
**						all the points in sp70_od, %48.
** 3.6 JGW 7/7/00 - 	[Borland C++] Further development of SP70 along the lines described
**						in sp70_od, %48. This version has been taken to the
**						stage where it can translate one alignment into encoded
**						patterns, unified patterns and an abstract pattern
**						(see sp70_od, %49).
** 3.7 JGW 18/7/00 -	[Borland C++] This version attempts to integrate parsing and learning
**						as described in sp70_od, %52. Development of the model
**						is finished at the point where it can create new
**						alignments from combinations of basic alignments
**						(see sp70_od, %53).
** 3.8 JGW 1/8/00 - 	[Borland C++] This version attempts to combine learning by combining
**						basic alignments with learning from partially-matched
**						patterns in New and Old (as described in sp70_od, %53).
** 4.0 JGW 29/9/00 -	[Visual C++] Re-creates functionality of v 3.8 (without
**						combine_sub_alignments()) but in the VC++ environment.
** 4.1 JGW 5/10/00 -	Developments as described in %56, %57.
** 4.2 JGW 18/10/00 -	Augmentation of combine_sub_alignments() as described in
**						sp70_od, %63 and %64. See also, %65, %66, %67 and %68.
** 4.3 JGW 28/10/00 -	See sp70_od, %69. This version attempts to
**						integrate the creation of new patterns and new
**						alignments that was developed in v 4.2 with
**						the procedure for forming such new patterns and
**						alignments from partial matches between
**						current_new_pattern and patterns in Old that
**						was developed in extract_patterns_and_classes().
**						As it turns out, this version merely bolts together,
**						the function combine_sub_alignments() and
**						extract_patterns_and_classes() (formerly
**						learn_new_patterns()) by running them one after
**						the other. The second function now includes
**						a test to see whether a given alignment is
**						suitable for processing by the function.
** 4.4 JGW 30/10/00 -	See sp70_od, %70, %72, %73. This version
**						can build a composite alignment, with associated
**						patterns and codes from a partial match between
**						New and a pattern in Old.
** 4.5 JGW 13/11/00 -	See sp70_od, %74, %75, %76, %77*. Development of
**						extract_patterns_and_classes() as a generalisation
**						of process_partly_matched_patterns(). The new function
**						extracts new patterns and classes from a partial
**						alignment and also isolates sub-alignments and
**						creates a new alignment from the sub-alignments
**						and the new patterns and classes.
** 4.6 JGW 2/1/01 -		Further development to find unmatched portions
**						of patterns and disjunctive classes, as described
**						in sp70_od, %78 (end) and %79.
** 4.7 JGW 11/1/01 -	Further development as described in sp70_od,
**						%81 and %82.
** 4.8 JGW 13/2/01 -	Testing of model on grammar induction proper
**						and further development to iron out problems.
**						(See SP70_od, %84, %85, %86, %87 and %88).
** 4.9 JGW 22/1/01 -	Development of the model to select the 'good'
**						patterns from amongst those produced during
**						learning. See sp70_od, %89.
** 5.0 JGW 8/3/01 -		Attempts to fix problems in sequencing of symbols
**						from current_new_pattern in recognise().
**						Incorporates write_alignment_vertical() from SP61
**						that writes alignments in vertical format. 
**						Continues development of ideas in sp70_od, %89,
**						following further thinking in %90. Incorporates
**						the method of measuring frequencies of occurrence
**						of patterns described in sp70_od, %92.
** 5.1 JGW 4/4/01 -		This model aims to tighten up measures of G and E
**						and ensure that the program seeks to minimise
**						(G + E). See sp70_od, %97 and %98. This version
**						attempts to implement the second of the solutions
**						proposed in %98 (see also %99).
** 5.2 JGW 18/4/01 -	Formation of a class cost_tree_object to hold
**						the classes hit_node and sequence.
** 5.3 JGW 18/4/01 -	Further development as described in sp70_od, 
**						%101 and %102. This version introduces an 
**						constraint on the formation of alternative
**						grammars: a limit on the total number of grammars
**						that may be considered at any one time. This
**						version also prints out values for G, E and
**						(G + E) for the best grammars found for 
**						successive patterns from New. The functions for
**						forming 'discrimination' symbols to distinguish
**						one member of a class from another have been
**						generalised so that a class may have any number
**						of members.
** 5.4 JGW 27/4/01 -	Further development and testing.
** 5.5 JGW 12/6/01 -	Reorganisation of extract_patterns_and_classes().
**						The program now rejects alignments for learning
**						new structures if it cannot find any valid
**						sub-alignments. The process for finding
**						sub-alignments and filling in classes between
**						them has been radically re-organised.
** 6.0 JGW 26/6/01 -	Further development to achieve 'correct'
**						grammars (see sp70_od, %112 ff.).
** 6.1 JGW 6/7/01 -		The first attempt to re-process one or
**						more grammars created by the program recursively
**						as discussed in sp70_od, %115. This version also
**						checks newly-created abstract patterns to make
**						sure there is not already a pattern with the same
**						CONTENTS symbols. If there is, that pattern is
**						used instead of the newly-created pattern. The 
**						program now contains a check, sar_array_is_OK(),
**						to make sure that it does not create spurious
**						composite alignments in which there are no new
**						unmatched sequences from pattern_new or pattern_old.
** 6.2 JGW 16/7/01 -	Further development as described in sp70_od, %121,
**						and %122.
** 6.3 JGW 17/7/01 -	Tries to get rid of 'inversion' of status of
**						symbols in New patterns.
** 6.4 JGW 17/7/01 -	Reorganisation of the assignment of class names
**						and discrimination symbols as outlined in
**						sp70_od, %123. This version runs slowly, apparently
**						because of the removal of restrictions on matching
**						IDENTIFICATION symbols against CONTENTS symbols
**						and vice vers and because there are now more
**						repeated discrimination symbols (see sp70_od,
**						%125).
** 6.5 JGW 19/7/01 -	Re-imposes restriction on matching, modified as
**						described in sp70_od, %125.
** 7.0 JGW 24/7/01 -	Introduces boundary markers for patterns and 
**						references and enhanced system for disjunctive
**						classes, sub-classes and cross-classification
**						(see sp70_od, %129). This development should
**						lead on to the reprocessing of newly-created
**						patterns 'on the fly' rather than all at the
**						end.
** 7.1 JGW 30/7/01 -	Attempts indexing of code symbols for cycles
**						other than first cycle (see sp70_od, %128).
** 7.2 JGW 2/8/01 -		The version of find_best_matches() and associated
**						functions that has been developed in SP61, v 6.2,
**						has been copied into this version of SP70.
** 7.3 JGW 8/8/01 -		Explores disjunctive sub-classes and
**						cross-classification (see sp70_od, %132 and %133).
** 9.0 JGW 4/9/01 -		Further development. Initially, the aim has been
**						to refine the process of checking newly-proposed
**						patterns against existing patterns by using
**						the existing capapility for partial matching
**						to yield sub-strucfure where appropriate 
**						(see sp70_od, %136). For the time being, the
**						program will be written up as it stands, with
**						any minor adjustments that may be needed.
** 9.1 JGW 6/11/01 -	Revision of method of calculating the frequencies
**						of symbol types (sp70_od, %137).
** 9.2 JGW 9/11/01 -	Attempts to correct error in construction of
**						abstract pattern described in sp70_od, %138.
*/

#include "SP70_head.h"

#define MARK_UNIFY_BEST FALSE // Switches on or off print routines which
	// mark the start and end of each call to unify_best.
#define PRINT_CD_VALUES FALSE
#define MAX_DR_PATTS_ALIGNMENT 10
#define CONTAINED_IN FALSE

char *in_filename ;
char *parameters_filename ;
char *out_filename ;
char *latex_filename ;
char *plot_filename ;
FILE *input_file ;
FILE *parameters_file ;
FILE *output_file ;
FILE *latex_file ;
FILE *plot_file ;

group *corpus ;
group *new_patterns ; // New patterns supplied at the start of processing
group *old_patterns ; // Old patterns supplied at the start of processing.
	// In 'learning' mode, there may not be any patterns in old_patterns.
group *list_of_combinations ; // Used in combine_sub_alignments.
group *original_symbols_in_corpus ;
group *symbols_in_old_after_learning ;
group *driving_patterns_and_alignments ; // For listing the alignments within the
	// keep_alignments which are to be used as driving patterns on
	// the next cycle.
group *keep_alignments ; // For listing the alignments from the
	// set of newly-formed alignments which are not to be purged.
group *current_alignments_from_parsing ; // This will hold newly-formed 
	// alignments from the current cycle.
group *full_alignments_from_parsing ; // Stores pointers to
	// 'full' alignments in which all CONTENTS symbols of Old patterns
	// are matched and all the symbols in current_new_pattern are matched.
group *global_alignments_from_parsing ; // At the end of each parsing cycle, 
	// alignments are moved from current_alignments_from_parsing
	// into global_alignments_from_parsing, leaving the first group
	// empty and ready for the next pattern from New.
group *created_alignments ; // Alignments created during learning.
group *created_code_patterns ;
group *set_of_class_records ; // Maintains a list of class_records,
	// without duplicates.
group *symbols_in_new ; // A list of the types of symbols (symbols with
	// a given name) found in new_patterns.
group *set_of_grammars, *set_of_cleaned_up_grammars ;
grammar *naive_grammar ; // Contains the patterns from New, each one
	// with added code symbols so that it can be used for encoding
	// data.
group *copies_of_grammar_patterns ;
class_record *current_class_record = NIL ;
sequence *current_new_pattern ;
hit_node *hit_root ; /* A pointer to the root of the hit structure */
hit_node *hn_master ; /* Used in add_hit(). */
int original_alphabet_size ; // The size of the alphabet of symbols 
	// used in the original corpus.
int current_alphabet_size ; // The size of the alphabet after learning
	// has been done, including the creation of new code symbols.
int group_ID_number = 1 ;
int sequence_ID_number = 1 ;
int hit_node_ID_number = 1 ;
int combination_ID_number = 1 ;
int grammar_ID_number = 1 ;
int class_record_ID_number = 1 ;

/*** ARRAYS, CONSTANTS AND INDICES FOR STORING SYMBOL COUNTS FOR
EACH SYMBOL IN NEW AND FOR STORING THE HIT STRUCTURE ***/

/*** STRUCTS, ARRAYS, CONSTANTS AND INDICES ***/

struct pnc_entry
{
	sequence *al1 ; // The alignment from which the pnc comes.
	sequence *pnc ; // A 2-row alignment which represents the pnc.
	int next_set ; // An index pointer to the next set of pncs.
	int next_pnc ; // An index pointer to the next pnc in
				// given set.
} ; // Used in probabilities_of_inferences

hit_node **leaf_array ; /* leaf_array[] is used for recording the
	leaf nodes of the hit structure. */

int *sort_array ; /* sort_array[] is used for sorting the
	entries in leaf_array[]. */

int fe_sort ; /* The first empty entry in sort_array[]. */

sequence **alignments_array ;

struct sub_alignment_entry *sub_alignments_array ; // Dynamic array
	// used in combine_sub_alignments.

int fe_sub_alignments_array ; // Used in combine_sub_alignments.

sequence *original_patterns_array[MEDIUM_SCRATCH_ARRAY_SIZE] ;

int half_hit_structure_rows ;

struct patterns_counts_record
{
	sequence *pattern1 ;
	int frequency, max_cioa, count_in_one_alignment ;
} ;

struct patterns_counts_record *patterns_counts_array ;

struct symbol_types_counts_record
{
	symbol *symbol_type ;
	int frequency, max_cioa, count_in_one_alignment ;
} ;

struct symbol_types_counts_record *symbol_types_counts_array ;

int size_of_old_patterns ;

struct sub_alignment_record
{
	list_element
		*el_pos_sa_start, *el_pos_sa_end, // Beginning and end of the
			// sub-alignment in the containing alignment.
		*el_pos_new_start, *el_pos_new_end, // Beginning and end
			// of hit sequence in pattern_new.
		*el_pos_old_start, *el_pos_old_end ; // Beginning and end
			// of hit sequence in pattern_old.
} sar_array[MEDIUM_SCRATCH_ARRAY_SIZE] ;

int fe_sar_array ;

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Checks a sequence of unmatched symbols within an alignment
**	to see whether or not it comprises a matching pair of brackets
**	with a context code symbol in between. If this test
**	succeeds, the corresponding class is identified and recorded
**	in the global variable current_class_record. This is used later
**	to ensure the assignment of the corresponding context
**	code symbols in the creation of any unmatched sequence
**	from current_new_sequence that lies 'opposite' the unmatched
**	symbols in Old. If the test succeeds, the function returns false,
**	otherwise it returns true.
**
**	This function is called from make_basic_pattern().
**
** CALLING SEQUENCE:
**
**	bool check_for_reference_pattern(sequence *pattern1,
**		sequence *basic_pattern)
**
** FORMAL ARGUMENTS:
**
**	Return value:		true if new code symbols are needed, 
**						false otherwise.
**	pattern1:			The pattern from which the unmatched symbols
**						were extracted.
**	basic_pattern:		The partially-constructed basic pattern
**						containing, at this stage, just the unmatched
**						symbols from Old.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	The function may assign a value to current_class_record.
**
** SIDE EFFECTS:
**
**	NONE
**
*/

bool check_for_reference_pattern(sequence *pattern1,
	sequence *basic_pattern)
{
	symbol *symbol1, *symbol2, *symbol3 ;
	list_element *el_pos1 ;

	if (pattern1 == current_new_pattern) return(true) ;

	// Check to see whether or not we have a matching pair
	// of brackets, with or without context code symbols in between.

	symbol1 = (symbol *)basic_pattern->get_first_child() ;
	symbol2 = (symbol *)basic_pattern->get_last_child() ;
	if (symbol1->get_type() != LEFT_BRACKET
		|| symbol2->get_type() != RIGHT_BRACKET) return(true) ;

	// Check that both symbols have status CONTENTS.

	if (symbol1->get_status() != CONTENTS) return(true) ;
	if (symbol2->get_status() != CONTENTS) return(true) ;

	// Check that there is one context code symbol between the
	// left and right brackets.

	int counter = 0 ;
	el_pos1 = basic_pattern->get_el_pos_by_child(symbol1) ;
	while (symbol3 = (symbol *)basic_pattern->
						get_next_child_by_el_pos(&el_pos1))
	{
		if (symbol3 == symbol2) break ;
		if (symbol3->get_type() != CONTEXT_CODE_SYMBOL) return(true) ;
		counter++ ;
	}

	if (counter > 1)
		abort_run("Too many code symbols between brackets.") ;

	return(false) ;

} // check_for_reference_pattern

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Called from extract_patterns_and_classes(). This function takes 
**	non-hit CONTENTS symbols from new_pattern or old_pattern and adds  
**	copies of them to encoded_new_pattern (or encoded_old_pattern).
**	It also adds an initial code symbol and discrimination symbol
**	at the beginning of the basic_pattern and brackets at beginning
**	and end.
**
** CALLING SEQUENCE:
**
**	sequence *make_basic_pattern(sequence *pattern1,
**		list_element *el_pos_start, list_element *el_pos_end)
**
** FORMAL ARGUMENTS:
**
**	Return value:		The newly-created basic pattern or NIL
**						if no such pattern can be formed.
**
**	pattern1:			The pattern which is to provide
**						the symbols to be copied into 
**						basic_pattern.
**	el_pos_start:		The first el_pos in the sequence to 
**						be encoded.
**	el_pos_end:			The last el_pos in the sequence to 
**						be encoded.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	The basic pattern is stored in old_patterns.
**	The details of the basic pattern are printed out.
**
** SIDE EFFECTS:
**
**	NONE
**
*/

sequence *make_basic_pattern(sequence *pattern1,
	list_element *el_pos_start, list_element *el_pos_end)
{
	list_element *pos1 ;
	symbol *symbol1, *copy_symbol, *ccs ;
	sequence *basic_pattern = new sequence(BASIC_PATTERN) ;
	class_record *class1 ;
	int counter = 0, int_pos_start = el_pos_start->get_position(),
		int_pos_end = el_pos_end->get_position() ;

	// Check that int_pos_start and int_pos_end represent an
	// actual sequence of symbols.

	if (int_pos_end - int_pos_start < 0) return(NIL) ;

	// Extract symbols from pattern1 between el_pos_start and
	// el_pos_end, inclusive.
	
	pos1 = el_pos_start ;
	
	for (symbol1 = (symbol *)pattern1->get_child_by_el_pos(&pos1);
		symbol1 != NIL ; symbol1 = (symbol *)
			pattern1->get_next_child_by_el_pos(&pos1))
	{
		if (symbol1->get_status() == IDENTIFICATION)
			break ; // End of CONTENTS symbols has been reached.

		copy_symbol = symbol1->shallow_copy() ;
		copy_symbol->set_status(CONTENTS) ;
		basic_pattern->add_child(copy_symbol) ;
		counter++ ;

		if (pos1 == el_pos_end) break ;
	}

	// Check to see that there is at least one symbol in the
	// sequence.

	if (counter < 1) return(NIL) ;

	// Examine the newly-extracted pattern to see whether it
	// comprises a matching pair of brackets with nothing but one
	// CONTEXT_CODE_SYMBOL between them. If yes, there is no need to make a new 
	// basic_pattern or new brackets and code symbols. Copies of the
	// existing code symbols will be used for any unmatched
	// pattern from New that lies 'opposite' the pair of code
	// symbols (in Old) that have just been found.

	if (check_for_reference_pattern(pattern1, basic_pattern)
		== false)
	{
		fprintf(output_file,
			"No basic pattern needed for the unmatched sequence ") ;
		basic_pattern->write_tree_object(PRINT_NO_FREQUENCIES) ;

		// Find the corresponding class and assign it to
		// current_class_record.

		ccs = basic_pattern->get_first_ccs() ; // There is no need
			// to check that there is more than one ccs because
			// this has already been checked in 
			// check_for_reference_pattern().
		if (ccs == NIL)
			abort_run("Anomalous ccs in make_basic_pattern()") ;
		class1 = find_class_record_by_name(ccs->get_name()) ;
		if (class1 == NIL)
			abort_run("Class not found in make_basic_pattern().") ;
		current_class_record = class1 ;

		delete basic_pattern ;
		fflush(output_file) ;
		return(NIL) ;
	}

	// Check to see whether there is any pre-existing pattern with
	// the same CONTENTS symbols.

	sequence *pre_existing_pattern = 
		basic_pattern->check_patterns() ;

	if (pre_existing_pattern == NIL)
	{
		old_patterns->add_child(basic_pattern) ;
		class1 = basic_pattern->add_ID_symbols(current_class_record) ;
		basic_pattern->compute_costs() ;

		fprintf(output_file, "NEW BASIC PATTERN: ") ;
		basic_pattern->print_ID() ;
		fprintf(output_file, ": ") ;
		basic_pattern->write_tree_object(PRINT_NO_FREQUENCIES) ;
		basic_pattern->write_with_details() ;
	}
	else
	{
		basic_pattern = pre_existing_pattern ;
		class1 = basic_pattern->add_ID_symbols(current_class_record) ;

		fprintf(output_file, "%s%d%s",
			"PRE-EXISTING BASIC PATTERN: ID",
			basic_pattern->get_ID(),
			": ") ;
		basic_pattern->write_tree_object(PRINT_NO_FREQUENCIES) ;
	}

	current_class_record = class1 ;

	return(basic_pattern) ;
} // make_basic_pattern

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Examines a portion of al1 between el_pos1 and el_pos2 (inclusive)
**	to see whether or not it matches any existing alignment. If it
**	does, the existing alignment is added to comb1. Otherwise the 
**	function makes a 'sub-alignment' from a partial alignment and
**	adds it comb1.
**
**	Before the sub-alignment is formed, a check is made to ensure
**	that it is valid. This means checking that all the symbols from
**	pattern_old within it are hit symbols and that the hit symbols
**	from New within the sub-alignment form a coherent sequence.
**
**	If the proposed sub-alignment is not valid, the function returns
**	false. Otherwise it returns true.
**
** CALLING SEQUENCE:
**
**	bool make_matched_sub_alignment(sequence *al1, 
**		int row_old, list_element *el_pos1, 
**		list_element *el_pos2, combination *comb1)
**
** FORMAL ARGUMENTS:
**
**	Return value:	true if we have a valid sub-alignment,
**					false otherwise.
**
**	al1:			The partial alignment.
**	row_old:		The row_old identified in
**						extract_patterns_and_classes().
**	el_pos1:		The el_pos of the first column of the
**						the sub-alignment in al1.
**	el_pos2:		The el_pos of the last column of the
**						the sub-alignment in al1.
**	comb1:			Stores the combination of sub-alignments.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	New sub-alignment stored in created
**
** SIDE EFFECTS:
**
**	NONE
**
*/

bool make_matched_sub_alignment(sequence *al1,
	int row_old, list_element *el_pos1, 
	list_element *el_pos2, combination *comb1)
{
	sequence *original_pattern ;
	symbol *original_symbol, *copy_symbol, *symbol1, *symbol2,
		*symbol3, *col1, *symbol_old, *symbol_new ;
	int i, j, sub_alignment_depth,
		al1_depth = al1->get_sequence_depth(),
		el_pos1_int_pos = el_pos1->get_position(),
		el_pos2_int_pos = el_pos2->get_position() ;
	list_element *el_pos3, *el_pos4 ;
	sequence *pattern_old = al1->get_row_pattern(row_old), 
		*new_abstract_pattern ;
	sub_alignment *new_sub_alignment ;
	list_element *pos3 ;

	// Check that original_patterns_array[] is big enough.

	if (al1_depth >= MEDIUM_SCRATCH_ARRAY_SIZE)
		abort_run("original_patterns_array[] is too small.") ;

	// Check that the proposed sub-alignment is valid. Every symbol
	// from pattern_old must be a hit symbol and, within the 
	// sequence, all the hit symbols from New must form a coherent
	// sequence. 

	int current_int_pos_new, previous_int_pos_new = NULL_VALUE,
		current_int_pos_old, previous_int_pos_old = NULL_VALUE ;

	el_pos4 = el_pos1 ;
	for (col1 = (symbol *)al1->get_child_by_el_pos(&el_pos4);
		col1 != NIL ; col1 = (symbol *)
			al1->get_next_child_by_el_pos(&el_pos4))
	{
		symbol_old = col1->get_row_symbol(row_old) ;
		if (symbol_old != NIL)
		{
			if (col1->is_a_hit() == false) return(false) ;
			current_int_pos_old = col1->get_row_orig_int_pos(row_old) ;

			if (previous_int_pos_old != NULL_VALUE
				&& current_int_pos_old - previous_int_pos_old > 1)
					return(false) ;
			previous_int_pos_old = current_int_pos_old ;
		}

		symbol_new = col1->get_row_symbol(0) ;
		if (symbol_new == NIL)
		{
			if (el_pos4 == el_pos2) break ; // The end of the proposed
				// sub-alignment has been reached.
			continue ;
		}

		if (col1->is_a_hit() == false) return(false) ; // Every
			// symbol from New within the alignment must be
			// a hit symbol.

		current_int_pos_new = col1->get_row_orig_int_pos(0) ;

		if (previous_int_pos_new != NULL_VALUE
			&& current_int_pos_new - previous_int_pos_new > 1)
				return(false) ;
		previous_int_pos_new = current_int_pos_new ;
		
		if (el_pos4 == el_pos2) break ; // The end of the proposed
			// sub-alignment has been reached.
	}

	// Find out how many rows there are in the sub-alignment.

	for (i = 0; i < al1_depth; i++) 
		original_patterns_array[i] = NIL ;

	el_pos3 = el_pos1 ;
	for (symbol1 = (symbol *)al1->get_child_by_el_pos(&el_pos3);
		symbol1 != NIL ; symbol1 = (symbol *)
			al1->get_next_child_by_el_pos(&el_pos3))
	{
		for (i = 0; i < al1_depth; i++)
		{
			symbol2 = symbol1->get_row_symbol(i) ;
			if (symbol2 != NIL)
				original_patterns_array[i] = 
					symbol1->get_row_pattern(i) ;
		}
		if (el_pos3 == el_pos2) break ;
	}

	sub_alignment_depth = 0 ;
	for (i = 0; i < al1_depth; i++)
	{
		if (original_patterns_array[i] != NIL) 
			sub_alignment_depth++ ;
	}

	int last_row = sub_alignment_depth - 1 ;

	// Match the sub-alignment against existing alignments. If an
	// existing alignment is found that matches the sub-alignment,
	// convert into class sub-alignment (by copying it into the new
	// type and deleting the original) and add the converted alignment
	// to comb1. Otherwise, create a new free-standing
	// alignment matching the sub-alignment (with new code symbols
	// and a new abstract pattern) and add it to comb1.

	sequence *test_alignment ;
	int sub_alignment_length = el_pos2_int_pos - el_pos1_int_pos + 1 ;	

	// In the following code, the depth of the sub-alignment is
	// given as 'sub_alignment_depth - 1' because, for the purpose
	// of matching with existing alignments, we exclude row_old from
	// consideration.

	class_record *class_ref1 ;
	sub_alignment *test_al1 ;
	list_for(test_alignment, sequence, current_alignments_from_parsing)
	{
		if (test_alignment->matches_sub_alignment(al1, al1_depth,
			sub_alignment_depth - 1, sub_alignment_length,
			el_pos1, el_pos2))
		{
			test_al1 = new sub_alignment(*test_alignment) ;
			current_alignments_from_parsing->
				extract_child(test_alignment) ;
			delete test_alignment ;
			current_alignments_from_parsing->add_child(test_al1) ;
			class_ref1 =
				find_class_record_from_pattern(test_al1) ;
			if (class_ref1 == NIL)
				abort_run("No class reference found in \
					make_matched_sub_alignment()") ;

			test_al1->set_class_reference(class_ref1) ;
			comb1->add_sub_alignment(test_al1, 
				el_pos1_int_pos, el_pos2_int_pos) ;

			return(true) ;
		}
	}

	list_for(test_alignment, sequence, created_alignments)
	{
		if (test_alignment->matches_sub_alignment(al1, al1_depth,
			sub_alignment_depth - 1, sub_alignment_length,
			el_pos1, el_pos2))
		{
			test_al1 = new sub_alignment(*test_alignment) ;
			created_alignments->extract_child(test_alignment) ;
			delete test_alignment ;
			current_alignments_from_parsing->add_child(test_al1) ;
			class_ref1 =
				find_class_record_from_pattern(test_al1) ;
			if (class_ref1 == NIL)
				abort_run("No class reference found in \
					make_matched_sub_alignment()") ;

			test_al1->set_class_reference(class_ref1) ;
			comb1->add_sub_alignment(test_al1, 
				el_pos1_int_pos, el_pos2_int_pos) ;

			return(true) ;
		}
	}

	// Create a free-standing copy of the sub-alignment with
	// a new abstract pattern.

	// Create a new abstract pattern with left and right brackets,
	// context symbol and discrimination symbol. NOTE: In cases where
	// al1 has only two rows, the 'abstract' pattern will be a pattern
	// in Old containing DATA_SYMBOLs flanked by brackets, with one or
	// more context code symbols and a discrimination symbol at the
	// beginning.

	new_abstract_pattern = new sequence(ABSTRACT_PATTERN) ;

	// Fill in 'abstract' symbols from row_old.

	el_pos3 = el_pos1 ;
	for (symbol1 = (symbol *)al1->get_child_by_el_pos(&el_pos3);
		symbol1 != NIL ; symbol1 = (symbol *)
			al1->get_next_child_by_el_pos(&el_pos3))
	{
		original_symbol = symbol1->get_row_symbol(row_old) ;
		if (original_symbol != NIL)
		{
			copy_symbol = new symbol(*original_symbol) ;
			new_abstract_pattern->add_child(copy_symbol) ;
		}
		if (el_pos3 == el_pos2) break ;
	}

	// Check whether there is any pre-existing pattern with all
	// and only the same CONTENTS symbols.

	sequence *existing_pattern = 
		new_abstract_pattern->check_patterns() ;

	class_record *class1 ;

	if (existing_pattern == NIL)
	{
		class1 = new_abstract_pattern->add_ID_symbols(NIL) ;

		// Store new_abstract_pattern and print out details.

		old_patterns->add_child(new_abstract_pattern) ;
		new_abstract_pattern->write_out_fully("NEW 'ABSTRACT' PATTERN:",
			NIL, NIL, write_section_chars_length, NULL_VALUE, true) ;
	}
	else
	{
		delete new_abstract_pattern ;
		new_abstract_pattern = existing_pattern ;
	}

	symbol *first_bracket_symbol = (symbol *)new_abstract_pattern->
							get_first_child() ;
	symbol *last_bracket_symbol = (symbol *)new_abstract_pattern->
							get_last_child() ;

	// Create a free-standing instance of the sub-alignment, add a copy
	// of the first_bracket_symbol and last_bracket_symbol and fill
	// in the columns.

	new_sub_alignment = new sub_alignment(SUB_ALIGNMENT) ;
	new_sub_alignment->set_sequence_depth(sub_alignment_depth) ;

	// Do first bracket column in the alignment.

	col1 = new symbol(first_bracket_symbol->get_name(),
					sub_alignment_depth, NULL_VALUE) ;
	col1->set_row_symbol(last_row, first_bracket_symbol) ;
	col1->set_row_pattern(0, current_new_pattern) ;
	col1->set_row_pattern(last_row, new_abstract_pattern) ;
	col1->set_sequence_depth(sub_alignment_depth) ;
	col1->set_status(IDENTIFICATION) ;
	col1->set_type(LEFT_BRACKET) ;
	col1->set_is_a_hit(false) ;
	new_sub_alignment->add_child(col1) ;

	// Fill in original_pattern for rows other than 
	// row 0 or row_old.

	j = 1 ;
	for (i = 1; i < al1_depth; i++)
	{
		if (i == row_old) continue ;
		original_pattern = original_patterns_array[i] ;
		if (original_pattern == NIL) continue ;
		col1->set_row_pattern(j, original_pattern) ;
		j++ ;
	}

	// Do last column in the alignment.

	col1 = new symbol(last_bracket_symbol->get_name(),
					sub_alignment_depth, NULL_VALUE) ;
	col1->set_row_symbol(last_row, last_bracket_symbol) ;
	col1->set_row_pattern(0, current_new_pattern) ;
	col1->set_row_pattern(last_row, new_abstract_pattern) ;
	col1->set_sequence_depth(sub_alignment_depth) ;
	col1->set_status(IDENTIFICATION) ;
	col1->set_type(RIGHT_BRACKET) ;
	col1->set_is_a_hit(false) ;
	new_sub_alignment->add_child(col1) ;
	symbol *terminal_column = col1 ;

	// Fill in original_pattern for rows other than 
	// row 0 or row_old.

	j = 1 ;
	for (i = 1; i < al1_depth; i++)
	{
		if (i == row_old) continue ;
		original_pattern = original_patterns_array[i] ;
		if (original_pattern == NIL) continue ;
		col1->set_row_pattern(j, original_pattern) ;
		j++ ;
	}

	// Find the first symbol in the new_abstract_pattern that is
	// not an IDENTIFICATION symbol and, at the same time, add
	// columns to the new sub-alignment corresponding to any new
	// code symbols that were added (above) by the
	// statement "new_abstract_pattern->add_ID_symbols()".

	list_element *pos2 = NIL ;
	while (symbol1 = (symbol *)
		new_abstract_pattern->get_next_child_by_el_pos(&pos2))
	{	
		if (symbol1->get_type() == IDENTIFICATION
			&& symbol1 != first_bracket_symbol)
		{
			col1 = new symbol(symbol1->get_name(),
					sub_alignment_depth, NULL_VALUE) ;

			col1->set_sequence_depth(sub_alignment_depth) ;
			col1->set_status(IDENTIFICATION) ;
			col1->set_type(symbol1->get_type()) ;
			col1->set_is_a_hit(false) ;
			col1->set_row_symbol(last_row, symbol1) ;
			col1->set_row_pattern(last_row, new_abstract_pattern) ;
			new_sub_alignment->precede(col1, terminal_column) ;
		}

		pos3 = pos2->get_next() ;
		symbol3 = (symbol *)pos3->get_el_obj() ;
		if (symbol3->get_status() != IDENTIFICATION) break ;
	}

	// Now fill in columns between the columns for first_bracket_symbol
	// and last_bracket_symbol.

	el_pos3 = el_pos1 ;
	for (symbol1 = (symbol *)al1->get_child_by_el_pos(&el_pos3);
		symbol1 != NIL ; symbol1 = (symbol *)
			al1->get_next_child_by_el_pos(&el_pos3))
	{
		col1 = new symbol(symbol1->get_name(),
					sub_alignment_depth, NULL_VALUE) ;

		col1->set_sequence_depth(sub_alignment_depth) ;
		col1->set_status(symbol1->get_status()) ;
		col1->set_type(symbol1->get_type()) ;
		col1->set_is_a_hit(symbol1->is_a_hit()) ;

		// Fill in details of 'abstract' symbol in last row of
		// the sub-alignment.

		original_symbol = symbol1->get_row_symbol(row_old) ;
		if (original_symbol != NIL)
		{
			copy_symbol = (symbol *)new_abstract_pattern->
							get_next_child_by_el_pos(&pos2) ;
			col1->set_row_symbol(last_row, copy_symbol) ;
		}

		col1->set_row_pattern(last_row, new_abstract_pattern) ;

		// Fill in details of symbol from New in first of
		// the sub-alignment.

		original_symbol = symbol1->get_row_symbol(0) ;
		if (original_symbol != NIL)
			col1->set_row_symbol(0, original_symbol) ;

		col1->set_row_pattern(0, current_new_pattern) ;

		// Set up the details of each of the remaining rows. 
		// The 'abstract' symbol (if any) has already been put in 
		// the last row and the symbol from New (if any) in the
		// first row. The details of other rows in symbol1 are 
		// entered in the same order as they appear in symbol1 
		// but without blank rows.

		j = 1 ;
		for (i = 1; i < al1_depth; i++)
		{
			if (i == row_old) continue ;
			if (original_patterns_array[i] == NIL) continue ;

			original_symbol = symbol1->get_row_symbol(i) ;
			if (original_symbol != NIL) 
				col1->set_row_symbol(j, original_symbol) ;
			original_pattern = symbol1->get_row_pattern(i) ;
			col1->set_row_pattern(j, original_pattern) ;
			j++ ;
		}

		col1->set_symbol_matches() ; 

		new_sub_alignment->precede(col1, terminal_column) ;

		if (el_pos3 == el_pos2) break ;
	}

	// Finish new_sub_alignment.

	new_sub_alignment->mark_parent_and_int_positions_non_recursive() ;

	// Fill in values for orig_patt_int_pos in each row of each column
	// of new_sub_alignment.

	alignment_element *al_el1 ;
	int orig_int_pos ;
	list_for(col1, symbol, new_sub_alignment)
	{
		for (i = 0; i < sub_alignment_depth; i++)
		{
			al_el1 = col1->get_al_el(i) ;
			original_symbol = (symbol *)al_el1->get_el_obj() ;
			if (original_symbol == NIL) continue ;
			original_pattern = al_el1->get_original_pattern() ;
			orig_int_pos = original_pattern->
				get_int_pos_by_child(original_symbol) ;
			al_el1->set_orig_patt_int_pos(orig_int_pos) ;
		}
	}

	// Make a code for the new alignment.

	sequence *new_code = new_sub_alignment->make_code(true) ;
	created_code_patterns->add_child(new_code) ;

	// Find the integer positions in current_new_pattern of the first
	// and last hit symbols in the sub-alignment.

	int first_int_pos_new = NULL_VALUE, last_int_pos_new ;

	list_for(col1, symbol, new_sub_alignment)
	{
		if (col1->is_a_hit() == false) continue ;
		symbol_new = col1->get_row_symbol(0) ;
		if (symbol_new == NIL) continue ;

		if (first_int_pos_new == NULL_VALUE)
			first_int_pos_new =
				current_new_pattern->get_int_pos_by_child(symbol_new) ;
		else last_int_pos_new =
				current_new_pattern->get_int_pos_by_child(symbol_new) ;
	}

	// Store new_sub_alignment and print out details.

	class_ref1 = find_class_record_from_pattern(new_sub_alignment) ;
	if (class_ref1 == NIL)
		abort_run("No class reference found in \
			make_matched_sub_alignment()") ;

	new_sub_alignment->set_class_reference(class_ref1) ;
	comb1->add_sub_alignment(new_sub_alignment, first_int_pos_new, 
		last_int_pos_new) ;
	new_sub_alignment->write_out_fully("NEW SUB-ALIGNMENT:",
		new_code, NIL, write_section_chars_length, NULL_VALUE, true) ;

	return(true) ;	

} // make_matched_sub_alignment

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	This function creates a new basic pattern from a portion of 
**	current_new_pattern (that may or may not have
**	been unified with a matching portion of a pattern from Old) and
**	incorporates this new basic pattern in a new alignment between
**	current_new_pattern and the newly-created basic pattern. The
**	new basic pattern is added to old_patterns and the
**	new alignment is added to created_alignments. The alignment
**	is also likely to be incorporated in a new composite alignment
**	(by the augment_set_of_combinations() function).
**
** CALLING SEQUENCE:
**
**	sub_alignment *make_unmatched_sub_alignment(sequence *new_basic_pattern,
**		list_element *el_pos1)
**
** FORMAL ARGUMENTS:
**
**	Return value:		A pointer to a new sub_alignment.
**
**	new_basic_pattern:	A newly-created pattern derived from part of
**						current_new_pattern.
**	el_pos1:			The position in current_new_pattern of the
**						first hit symbol in the alignment.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

sub_alignment *make_unmatched_sub_alignment(sequence *new_basic_pattern,
	list_element *el_pos1)
{
	sub_alignment *new_sub_alignment = new sub_alignment(SUB_ALIGNMENT) ;
	symbol *symbol1, *new_column, *symbol_from_new ;
	list_element *el_pos3 = el_pos1 ;

	new_sub_alignment->set_sequence_depth(2) ;

	symbol_from_new = (symbol *)
		current_new_pattern->get_child_by_el_pos(&el_pos3) ;

	created_alignments->add_child(new_sub_alignment) ;

	list_for(symbol1, symbol, new_basic_pattern)
	{
		new_column = new symbol(symbol1->get_name(), 2, NULL_VALUE) ;
		new_sub_alignment->add_child(new_column) ;
		new_column->set_status(symbol1->get_status()) ;
		new_column->set_type(symbol1->get_type()) ;
		new_column->set_bit_cost(symbol1->get_bit_cost()) ;
		new_column->set_row_symbol(1, symbol1) ;
		new_column->set_row_pattern(0, current_new_pattern) ;
		new_column->set_row_pattern(1, new_basic_pattern) ;
		new_column->set_row_orig_int_pos(1, 
				new_basic_pattern->get_int_pos_by_child(symbol1)) ;
		if (symbol1->get_status() == IDENTIFICATION) continue ;

		// new_column represents a hit between a symbol in New and
		// a matching symbol in new_basic_pattern.

		new_column->set_is_a_hit(true) ;

		// Add symbol_from_new to the column in row 0, together 
		// with other details for that row. Then find the next
		// value for symbol_from_new, ready for the next cycle.

		new_column->set_row_symbol(0, symbol_from_new) ;
		new_column->set_row_orig_int_pos(0, el_pos3->get_position()) ;

		new_column->set_symbol_matches() ; // Marks the matches
			// between symbols from top to bottom of the row.

		symbol_from_new = (symbol *)
			current_new_pattern->get_next_child_by_el_pos(&el_pos3) ;
	}

	new_sub_alignment->mark_parent_and_int_positions_non_recursive() ;

	// Print out new_sub_alignment both as alignment and
	// as a flat pattern, with and without details.

	sequence *code = new_sub_alignment->make_code(true) ;
	created_code_patterns->add_child(code) ;

	new_sub_alignment->write_out_fully("NEW BASIC ALIGNMENT",
		NIL, NIL, write_section_chars_length, NULL_VALUE, true) ;

	return(new_sub_alignment) ;

} // make_unmatched_sub_alignment

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Called from extract_patterns_and_classes().
**
**	Makes a new basic pattern from an unmatched portion of 
**	current_new_pattern, incorporates this in a new basic alignment
**	and adds the basic alignment to a combination of such alignments
**	for current_new_pattern.
**
** CALLING SEQUENCE:
**
**	void make_pattern_and_alignment(sequence *pattern1,
**		list_element *el_pos_start, list_element *el_pos_end,
**		combination *comb1)
**
** FORMAL ARGUMENTS:
**
**	Return value:		void
**
**	pattern1:			The pattern from which the basic pattern
**						is to be extracted (normally
**						current_new_pattern).
**	el_pos_start:		The first el_pos in the unmatched portion
**						of current_new_pattern.
**	el_pos_end:			The last el_pos in the unmatched portion
**						of current_new_pattern.
**	comb1:				The combination being built.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	New pattern, alignment and augmentation of comb1
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void make_pattern_and_alignment(sequence *pattern1,
	list_element *el_pos_start, list_element *el_pos_end,
	combination *comb1)
{
	int int_pos1, int_pos2 ;
	sub_alignment *sub_al1 ;

	sequence *basic_pattern = 
		make_basic_pattern(pattern1,
			el_pos_start, el_pos_end) ;

	if (basic_pattern == NIL)
	{
		fprintf(output_file, "%s%s%d%s",
			"No new basic pattern (or corresponding alignment)\n",
			"formed from pattern1 in ID",
			pattern1->get_ID(),
			"\n\n") ;
		return ;
	}
	else sub_al1 = 
		make_unmatched_sub_alignment(basic_pattern, el_pos_start) ;

	int_pos1 = el_pos_start->get_position() ;
	int_pos2 = el_pos_end->get_position() ;

	sub_al1->set_class_reference(current_class_record) ;
	comb1->add_sub_alignment(sub_al1, int_pos1, 
		int_pos2) ;
} // make_pattern_and_alignment

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Starting from el_pos1, this function searches through al1 looking
**	for a valid sub-alignment. This is a sequence of columns within
**	al1 where every column that has a symbol in row_old is also a hit
**	column and where all hit symbols from pattern_new within the sequence
**	form a coherent sequence within pattern_new.
**
**	If the function meets a situation which shows that al1 cannot be
**	divided into sub-alignments (with unmatched sequences in pattern_new
**	and pattern_old between them), then *al_pos_end_sa is given the value
**	NIL. If the function fails to find a sub-alignment but without
**	evidence that al1 cannot be divided into sub-alignments, then
**	*al_pos_end_sa is given the same value as el_pos1.
**
** CALLING SEQUENCE:
**
**	bool find_sub_alignment(sequence *al1, list_element *el_pos1,
**		sequence *pattern_new, sequence *pattern_old, int row_old,
**		bool *alignment_is_OK)
**
** FORMAL ARGUMENTS:
**
**	Return value:			true if a valid sub-alignment has been 
**							found, false otherwise.
**
**	al1:					The alignment within which a sub-alignment
**							is sought.
**	el_pos1:				The position within al1 where searching
**							is to start.
**	pattern_new:			The current pattern from New.
**	pattern_old:			pattern_old within al1.
**	row_old:				The row within al1 containing pattern_old.
**	alignment_is_OK:		A pointer to a boolean recording whether
**							or not a valid sub-alignment has been
**							found.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

bool find_sub_alignment(sequence *al1, list_element *el_pos1,
	sequence *pattern_new, sequence *pattern_old, int row_old,
	bool *alignment_is_OK)
{
	symbol *al_symbol, *symbol_new, *symbol_old, *symbol1 ;
	int symbol_new_current_int_pos,
		symbol_new_previous_int_pos = NULL_VALUE,
		al1_depth = al1->get_sequence_depth(),
		row_counter ;
	bool at_least_one_hit_symbol_in_new_found = false,
		current_column_is_grounded, previous_column_is_grounded ; // A
			// column within al1 is 'grounded' if it contains a
			// symbol from pattern_old and if the column is a hit column.
	list_element
		*el_pos_sa_start = el_pos1, *el_pos_sa_end, // Beginning and end 
			// of the sub-alignment in the containing alignment.
		*el_pos_new_start = NIL, *el_pos_new_end, // Beginning and end
			// of hit sequence in pattern_new.
		*el_pos_old_start = NIL, *el_pos_old_end ; // Beginning and end
			// of hit sequence in pattern_old.

	// Find the start of a sub-alignment.

	for (al_symbol = (symbol *)al1->get_next_child_by_el_pos(&el_pos_sa_start);
		al_symbol != NIL; al_symbol = (symbol *)
			al1->get_next_child_by_el_pos(&el_pos_sa_start))
	{
		symbol_old = al_symbol->get_row_symbol(row_old) ;
		if (symbol_old == NIL) continue ;
		if (al_symbol->is_a_hit() == false
			|| symbol_old->get_status() != CONTENTS) continue ;

		// We have found a valid symbol in pattern_old for the
		// start of a sub-alignment.

		el_pos_old_start = pattern_old->get_el_pos_by_child(symbol_old) ;

		break ;
	}

	if (al_symbol == NIL)
	{
		*alignment_is_OK = true ;
		return(false) ; // The end of al1 has been
			// reached without finding the start of a sub-alignment.
	}

	// Now search for the end of the alignment, checking that there
	// is a coherent sequence of hit symbols in pattern_new and
	// recording the start and end positions of that sequence.
	// The sub-alignment ends when symbols in pattern_old no longer
	// form a coherent hit sequence or when they are no longer
	// CONTENTS symbols. The sub-alignment also ends if any non-hit
	// symbol from pattern_new is found provided that this break
	// occurs in a column immediately following one containing a 
	// a valid hit symbol in pattern_old. If a break in the sequence
	// of hit symbols in pattern_new occurs if this condition does
	// not obtain, then the sub-alignment is not valid.
	// All other CONTENTS symbols in the alignment must also be
	// hit symbols. The rules are summarised in SP70_od, %113.

	previous_column_is_grounded = false ;
	list_element *el_pos2 = el_pos_sa_start ;
	symbol_new_current_int_pos = NULL_VALUE ;
	for (al_symbol = (symbol *)al1->get_child_by_el_pos(&el_pos2);
		al_symbol != NIL; al_symbol = (symbol *)
			al1->get_next_child_by_el_pos(&el_pos2))
	{
		symbol_old = al_symbol->get_row_symbol(row_old) ;

		symbol_new = al_symbol->get_row_symbol(0) ;

		if (symbol_new != NIL)
		{
			if (symbol_new_current_int_pos == NULL_VALUE)
				el_pos_new_start =
					pattern_new->get_el_pos_by_child(symbol_new) ;

			symbol_new_current_int_pos =
				al_symbol->get_row_orig_int_pos(0) ;

			if (symbol_new_previous_int_pos != NULL_VALUE)
			{
				if (symbol_new_current_int_pos -
					symbol_new_previous_int_pos > 1)
				{
					if (at_least_one_hit_symbol_in_new_found
						&& previous_column_is_grounded)
					{
						*alignment_is_OK = true ;
						break ;
					}
					else
					{
						*alignment_is_OK = false ;
						return(false) ;
					}
				}
			}

			if (at_least_one_hit_symbol_in_new_found == false)
				at_least_one_hit_symbol_in_new_found = true ;

			el_pos_new_end = pattern_new->
				get_el_pos_by_child(symbol_new) ; // If the end of
				// the sub-alignment is found on the next iteration,
				// this value will be valid for the end of
				// the sub-alignment.

			symbol_new_previous_int_pos = symbol_new_current_int_pos ;
		}

		if (symbol_old != NIL)
		{
			if (al_symbol->is_a_hit() == false
				|| symbol_old->get_status() != CONTENTS)
			{
				if (at_least_one_hit_symbol_in_new_found)
				{
					*alignment_is_OK = true ;
					break ;
				}
				else
				{
					*alignment_is_OK = false ;
					return(false) ;
				}
			}

			current_column_is_grounded = true ;
			el_pos_old_end = pattern_old->
				get_el_pos_by_child(symbol_old) ;
			el_pos_sa_end = el_pos2 ; // If the end of
				// the sub-alignment is found on the next iteration,
				// this value (and the value of el_pos_old_end) will
				// be valid for the end of the sub-alignment.
		}
		else current_column_is_grounded = false ;

		// Check whether there are any other CONTENTS symbols in 
		// this column. If there are, check that the column is
		// a hit column. If it is not, the sub-alignment is not
		// valid.

		if (al1_depth > 2)
		{
			for (row_counter = 1; row_counter < al1_depth; row_counter++)
			{
				if (row_counter == row_old) continue ;

				symbol1 = al_symbol->get_row_symbol(row_counter) ;
				if (symbol1 == NIL) continue ;

				if (symbol1->get_status() != CONTENTS) continue ;

				if (al_symbol->is_a_hit() == false)
				{
					*alignment_is_OK = false ;
					return(false) ;
				}
			}
		}

		previous_column_is_grounded = current_column_is_grounded ;
	}

	sar_array[fe_sar_array].el_pos_sa_start = el_pos_sa_start ;	
	sar_array[fe_sar_array].el_pos_sa_end = el_pos_sa_end ;	
	sar_array[fe_sar_array].el_pos_new_start = el_pos_new_start ;	
	sar_array[fe_sar_array].el_pos_new_end = el_pos_new_end ;	
	sar_array[fe_sar_array].el_pos_old_start = el_pos_old_start ;	
	sar_array[fe_sar_array].el_pos_old_end = el_pos_old_end ;	

	return(true) ;

} // find_sub_alignment

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Checks sar_array[] to make sure that there is at least one
**	unmatched sequence of CONTENTS symbols in pattern_old or pattern_new,
**	or both. If no such unmatched sequences are found,
**	There is no case for making new sub-alignments or a new
**	composite alignment.
**
** CALLING SEQUENCE:
**
**	bool sar_array_is_OK(sequence *pattern_new, sequence *pattern_old,
**		list_element *el_pos_new_start, list_element *el_pos_old_start)
**
** FORMAL ARGUMENTS:
**
**	Return value:		true if the alignment contains unmatched patterns,
**						false otherwise.
**
**	pattern_new:		The current pattern from New in the alignment being
**						processed.
**	pattern_old:		The current pattern from Old in the alignment being
**						processed.
**	el_pos_new_start:	The position of the first CONTENTS symbol in
**						pattern_new.
**	el_pos_old_start:	The position of the first CONTENTS symbol in
**						pattern_old.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

bool sar_array_is_OK(sequence *pattern_new, sequence *pattern_old,
	list_element *el_pos_new_start, list_element *el_pos_old_start)
{
	list_element *epn_start = el_pos_new_start,
		*epo_start = el_pos_old_start, *epn_end, *epo_end ;
	int int_pos_start, int_pos_end, index1 ;

	for (index1 = 0; index1 < fe_sar_array; index1++)
	{
		// Look at pattern_new.

		epn_end = pattern_new->
			get_preceding_el_pos(sar_array[index1].el_pos_new_start) ;

		if (epn_end != NIL)
		{
			int_pos_start = epn_start->get_position() ;
			int_pos_end = epn_end->get_position() ;

			if (int_pos_end - int_pos_start >= 0)
				return(true) ;
		}

		// Look at pattern_old.

		epo_end = pattern_old->
			get_preceding_el_pos(sar_array[index1].el_pos_old_start) ;

		if (epo_end != NIL)
		{
			int_pos_start = epo_start->get_position() ;
			int_pos_end = epo_end->get_position() ;

			if (int_pos_end - int_pos_start >= 0)
				return(true) ;
		}

		// Find values for epn_start and epo_start
		// ready for the next iteration.

		epn_start = pattern_new->
			get_next_el_pos(&(sar_array[index1].el_pos_new_end)) ;
		epo_start = pattern_old->
			get_next_el_pos(&(sar_array[index1].el_pos_old_end)) ;
	}

	// Now do the tail end of the array.

	symbol *symbol1 ;
	if (epn_start != NIL)
	{
		// Find the last symbol in pattern_new that is a
		// CONTENTS symbol.

		epn_end = pattern_new->get_last_el_pos() ;

		symbol1 = (symbol *)epn_end->get_el_obj() ;
		while (symbol1->get_status() != CONTENTS)
		{
			epn_end = pattern_new->
				get_preceding_el_pos(epn_end) ;
			symbol1 = (symbol *)epn_end->get_el_obj() ;
		}

		int_pos_start = epn_start->get_position() ;
		int_pos_end = epn_end->get_position() ;

		if (int_pos_end - int_pos_start >= 0)
			return(true) ;
	}

	if (epo_start != NIL)
	{
		// Find the last symbol in pattern_old that is a
		// CONTENTS symbol.

		epo_end = pattern_old->get_last_el_pos() ;

		symbol1 = (symbol *)epo_end->get_el_obj() ;
		while (symbol1->get_status() != CONTENTS)
		{
			epo_end = pattern_old->
				get_preceding_el_pos(epo_end) ;
			symbol1 = (symbol *)epo_end->get_el_obj() ;
		}

		int_pos_start = epo_start->get_position() ;
		int_pos_end = epo_end->get_position() ;

		if (int_pos_end - int_pos_start >= 0)
			return(true) ;
	}

	return(false) ;

} // sar_array_is_OK

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	This function processes al_constr to find unmatched portions of 
**	current_new_pattern in the alignment.
**
**	If there are unmatched portion(s) of New, process the alignment
**	as described in sp70_od, %76 to extract new patterns and classes
**	(where appropriate) and make a new alignment containing those
**	patterns and classes.
**
** CALLING SEQUENCE:
**
**	void extract_patterns_and_classes(sequence *al1)
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
**	al1:			The alignment to be processed.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	Newly-created patterns (added to old_patterns) and new alignments
**	(added to created_alignments).
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void extract_patterns_and_classes(sequence *al1)
{
	int length_cnp = current_new_pattern->get_number_of_children(), i ;
	combination *comb1 ;
	symbol *symbol1, *al_symbol ;
	sequence *pattern_old, *pattern_new, *basic_pattern_old,
		*unified_pattern ;

	// 1 Preliminaries.

	fprintf(output_file, "%s%d%s", 
		"Start of extract_patterns_and_classes() (ID",
		al1->get_ID(),
		")\n\n") ;

	comb1 = new combination(length_cnp) ; // This is the combination
		// that will store the basic alignments created 
		// in this function.

	// PRINTING

	fprintf(output_file, "FROM ALIGNMENT ") ;
	al1->print_ID() ;
	fprintf(output_file, "\n\n") ;
	al1->write_alignment(output_file, write_section_chars_length, 
		NULL_VALUE, alignment_format) ;
	fprintf(output_file,
		"IS FORMED THESE CODED PATTERNS AND ALIGNMENTS:\n\n") ;
	fflush(output_file) ;

 	// The values of the following three patterns are set to NIL
	// because there are places where it is necessary to test
	// whether or not one of these patterns has been formed.

	basic_pattern_old = unified_pattern = NIL ;

	// 2 Look for the most abstract pattern in the alignment. This
	// is the one that starts furthest to the left. In cases where
	// an alignment has only two rows, the most abstract pattern may
	// not be very abstract. Hence, the most abstract pattern found
	// is designated pattern_old rather than something like
	// 'abstract_pattern' (and the corresponding row is row_old).

	int al1_depth = al1->get_sequence_depth(), row_old ;

	al_symbol = (symbol *)al1->get_first_child() ;
	for (i = 1; i < al1_depth; i++) // Skip over row 0.
	{
		symbol1 = al_symbol->get_row_symbol(i) ;
		if (symbol1 != NIL)
		{
			pattern_old = al_symbol->get_row_pattern(i) ;
			row_old = i ;
			break ;
		}
	}

	// 3 Scan al1 looking for 'sub-alignments', meaning parts
	// of al1 that can be free-standing alignments that are
	// rooted on pattern_old. At the same time, the system looks
	// for unmatched portions of current_new_pattern and unmatched
	// portions of pattern_old. These are copied into 'encoded' patterns
	// each with their own code symbols. If a basic pattern from
	// current_new_pattern lies 'opposite' a basic pattern from
	// pattern_old, then they both have the same class symbols but
	// each has a distinctive discrimination symbol.

	// In the following loop, we are looking for 'sub-alignments' and
	// gaps between sub-alignments. 

	// A sub-alignment is a sequence of columns within the alignment
	// where all the symbols in pattern_old are hit symbols that form 
	// a coherent sequence and where all the symbols from New that
	// appear within that sequence of columns are hit symbols that form
	// a coherent sequence.

	// Between one sub-alignment and another, or between a
	// sub-alignment and the start or finish of the alignment,
	// there may be a coherent sequence of non-hit symbols from
	// new or a coherent sequence of non-hit symbols from pattern_old.
	// Either or both of these sequences may be formed into a
	// new basic pattern. Where two basic patterns lie opposite
	// each other within the original alignment, they may be
	// formed into a disjunctive class.

	// Start by identifying valid sub-alignments and storing their
	// details. Then portions of pattern_new and pattern_old that lie
	// between sub-alignments or between a sub-alignment and the
	// start or end of an alignment can be identified.

	list_element *el_pos1 = NIL ; // The position in al1 where the search
		// for a sub-alignment begins.
	bool alignment_is_OK ;

	pattern_new = al1->get_row_pattern(0) ;
	fe_sar_array = 0 ;

	while (find_sub_alignment(al1, el_pos1, pattern_new, pattern_old,
		row_old, &alignment_is_OK))
	{
		el_pos1 = sar_array[fe_sar_array].el_pos_sa_end ;
		increment_index(&fe_sar_array, MEDIUM_SCRATCH_ARRAY_SIZE,
			"Overflow of sar_array[]") ;
	}

	if (alignment_is_OK == false // An invalid sub-alignment was found.
		|| fe_sar_array == 0) // No sub-alignment was found.
	{
		fprintf(output_file, "%s%d%s",
			"No sub-alignment or invalid sub-alignment in alignment ID",
			al1->get_ID(),
			". This alignment is abandoned.\n\n") ;

		delete comb1 ;

		return ;
	}

	// Now check sub-alignments to see if any already exist. Otherwise,
	// create new sub-alignments and store them in comb1.
	// At the same time, check the unmatched sections of pattern_new
	// and pattern_old and create new patterns and alignments
	// where appropriate.

	int index1, int_pos_start, int_pos_end ;
	list_element *el_pos_sa_start, *el_pos_sa_end,
		*el_pos_new_start, *el_pos_new_end, *el_pos_old_start,
		*el_pos_old_end ;
	sequence *basic_pattern ;

	// Find position of first CONTENTS symbol in pattern_new.

	el_pos_new_start = pattern_new->get_first_el_pos() ;
	symbol1 = (symbol *)el_pos_new_start->get_el_obj() ;
	while (symbol1->get_status() == IDENTIFICATION)
	{
		el_pos_new_start = el_pos_new_start->get_next() ;
		symbol1 = (symbol *)el_pos_new_start->get_el_obj() ;
	}

	// Find position of first CONTENTS symbol in pattern_old.

	el_pos_old_start = pattern_old->get_first_el_pos() ;
	symbol1 = (symbol *)el_pos_old_start->get_el_obj() ;
	while (symbol1->get_status() == IDENTIFICATION)
	{
		el_pos_old_start = el_pos_old_start->get_next() ;
		symbol1 = (symbol *)el_pos_old_start->get_el_obj() ;
	}

	// Check sar_array[] to make sure that there is at least one
	// unmatched sequence of CONTENTS symbols in pattern_old or in
	// pattern_new, or both. If no such unmatched sequences are found,
	// There is no case for making new sub-alignments or a new
	// composite alignment.

	if (sar_array_is_OK(pattern_new, pattern_old,
		el_pos_new_start, el_pos_old_start) == false)
	{
		fprintf(output_file, "%s%d%s",
			"No new patterns to be formed from alignment ID",
			al1->get_ID(),
			". This alignment is abandoned.\n\n") ;

		delete comb1 ;

		return ;
	}

	// Now process sar_array[].

	for (index1 = 0; index1 < fe_sar_array; index1++)
	{
		current_class_record = NIL ; // This ensures that the
			// first or only unmatched portion of New or Old that
			// that is made into a new pattern, is assigned to
			// a new class, with appropriate IDENTIFICATION symbols.

		el_pos_sa_start = sar_array[index1].el_pos_sa_start ;
		el_pos_sa_end = sar_array[index1].el_pos_sa_end ;

		make_matched_sub_alignment(al1, row_old, el_pos_sa_start,
			el_pos_sa_end, comb1) ;

		// Check to see whether there is an unmatched portion of
		// pattern_old before the present sub-alignment. If there
		// is, create a new pattern from this portion.

		el_pos_old_end = pattern_old->
			get_preceding_el_pos(sar_array[index1].el_pos_old_start) ;

		if (el_pos_old_end == NIL)
		{
			fprintf(output_file, "%s%d%s",
				"No new basic pattern formed from pattern_old in ID",
				al1->get_ID(),
				"\nat the start of the alignment\n\n") ;
		}
		else
		{
			int_pos_start = el_pos_old_start->get_position() ;
			int_pos_end = el_pos_old_end->get_position() ;

			if (int_pos_end - int_pos_start >= 0)
			{
				basic_pattern = make_basic_pattern(pattern_old,
					el_pos_old_start, el_pos_old_end) ;

				if (basic_pattern == NIL)
				{
					fprintf(output_file, "%s%d%s%d%s%d%s",
						"No new basic pattern formed from pattern_old in ID",
						al1->get_ID(),
						"\nbetween ",
						int_pos_start,
						" and ",
						int_pos_end,
						" inclusive.\n\n") ;
				}
			}
		}

		// Check to see whether there is an unmatched portion of
		// pattern_new before the present sub-alignment. If there
		// is, create a new pattern and corresponding alignment and
		// add the alignment to comb1. First, find a value for
		// el_pos_new_end for this unmatched portion of pattern_new.

		el_pos_new_end = pattern_new->
			get_preceding_el_pos(sar_array[index1].el_pos_new_start) ;

		// Check that there is at least one symbol in the pattern
		// to be formed.

		if (el_pos_new_end == NIL)
		{
			fprintf(output_file, "%s%d%s",
				"No new basic pattern formed from pattern_new in ID",
				al1->get_ID(),
				"\nat the start of the alignment\n\n") ;			
		}
		else
		{
			int_pos_start = el_pos_new_start->get_position() ;
			int_pos_end = el_pos_new_end->get_position() ;

			if (int_pos_end - int_pos_start >= 0)
				make_pattern_and_alignment(pattern_new, el_pos_new_start,
					el_pos_new_end, comb1) ;
		}

		// Find values for el_pos_new_start and el_pos_old_start
		// ready for the next iteration.

		el_pos_new_start = pattern_new->
			get_next_el_pos(&(sar_array[index1].el_pos_new_end)) ;
		el_pos_old_start = pattern_old->
			get_next_el_pos(&(sar_array[index1].el_pos_old_end)) ;
	}

	// Check tail end of the alignment.

	current_class_record = NIL ; 

	if (el_pos_old_start != NIL)
	{
		// Find the last symbol in pattern_old that is a CONTENTS symbol.

		el_pos_old_end = pattern_old->get_last_el_pos() ;

		symbol1 = (symbol *)el_pos_old_end->get_el_obj() ;
		while (symbol1->get_status() != CONTENTS)
		{
			el_pos_old_end = pattern_old->
				get_preceding_el_pos(el_pos_old_end) ;
			symbol1 = (symbol *)el_pos_old_end->get_el_obj() ;
		}

		basic_pattern = make_basic_pattern(pattern_old,
			el_pos_old_start, el_pos_old_end) ;

		if (basic_pattern == NIL)
		{
			int_pos_start = el_pos_old_start->get_position() ;
			int_pos_end = el_pos_old_end->get_position() ;
			fprintf(output_file, "%s%d%s%d%s%d%s",
				"No new basic pattern formed from pattern_old in ID",
				al1->get_ID(),
				"\nbetween ",
				int_pos_start,
				" and ",
				int_pos_end,
				"\n\n") ;
		}
	}

	if (el_pos_new_start != NIL)
	{
		// Find the last symbol in pattern_new that is a
		// CONTENTS symbol.

		el_pos_new_end = pattern_new->get_last_el_pos() ;

		symbol1 = (symbol *)el_pos_new_end->get_el_obj() ;
		while (symbol1->get_status() != CONTENTS)
		{
			el_pos_new_end = pattern_new->
				get_preceding_el_pos(el_pos_new_end) ;
			symbol1 = (symbol *)el_pos_new_end->get_el_obj() ;
		}

		make_pattern_and_alignment(pattern_new, el_pos_new_start,
			el_pos_new_end, comb1) ;		
	}

	comb1->make_composite_alignment() ; // Convert the combination
		// into a composite aligment, including an abstract pattern
		// to tie the basic alignments together.

	delete comb1 ;

} // extract_patterns_and_classes

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Called from combine_sub_alignments. This function systematically
**	forms combinations of basic alignments and scores them.
**
** CALLING SEQUENCE:
**
**	void combine_alignments(int ba_index, combination *comb1)
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
**	ba_index:		An index into sub_alignments_array[], that stores
**					the list of basic alignments which are to be combined.
**
**	combination:	A list of entries showing basic alignments that
**					fit together against the current pattern from New,
**					without any overlap. At the top level of this recursive
**					function, combination is empty.
**
** IMPLICIT INPUTS:
**
**	sub_alignments_array[]
**
** IMPLICIT OUTPUTS:
**
**	A list of combinations of basic alignments.
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void combine_alignments(int ba_index, combination *comb1)
{
	combination *comb2 ;
	struct sub_alignment_entry *basic_al_entry ;
	int ba_index1, pos_new_first, pos_new_last ;
	sub_alignment *basic_al1 ;
	class_record *class_ref1 ;
	
	for (ba_index1 = ba_index + 1; 
		ba_index1 < fe_sub_alignments_array; ba_index1++)
	{
		basic_al_entry = &sub_alignments_array[ba_index1] ;
		basic_al1 = basic_al_entry->sub_al1 ;
		pos_new_first = basic_al_entry->pos_new_first ;
		pos_new_last = basic_al_entry->pos_new_last ;
		if (comb1->can_accept(pos_new_first, pos_new_last))
		{
			comb2 = new combination(*comb1) ; // Make copy of comb1.

			class_ref1 =
				find_class_record_from_pattern(basic_al1) ;
			if (class_ref1 == NIL)
				abort_run("No class reference found in \
					combine_alignments()") ;

			basic_al1->set_class_reference(class_ref1) ;
			comb2->add_sub_alignment(basic_al1,
				pos_new_first, pos_new_last) ;
			list_of_combinations->add_child(comb2) ;
			combine_alignments(ba_index1, comb2) ;	
		}
		else combine_alignments(ba_index1, comb1) ;
	}
} // combine_alignments

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Sorts a list of combinations into descending order of
**	combination_score.
**
**	For reasons of efficiency, sorting is done from lowest to
**	highest score. Otherwise, the method has been chosen because
**	it is simple to program rather than particularly efficient.
**
** CALLING SEQUENCE:
**
**	void sort_combinations()
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	The sorted list of combinations.
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void sort_combinations()
{
	group *temp_list = new group ;
	combination *comb1, *worst_combination ;
	double combination_score, lowest_combination_score ;
	
	while (true)
	{
		lowest_combination_score = VERY_LARGE_NUMBER ;
		worst_combination = NIL ;
		for (comb1 = (combination *)list_of_combinations->
			get_first_child(); comb1 != NIL;
		comb1 = (combination *)list_of_combinations->
			get_next_child())
		{
			combination_score = comb1->get_combination_score() ;
			if (combination_score < lowest_combination_score)
			{
				lowest_combination_score = combination_score ;
				worst_combination = comb1 ;
			}
		}
		
		if (worst_combination == NIL) break ;
		
		list_of_combinations->extract_child(worst_combination) ;
		temp_list->add_child_at_start(worst_combination) ;
	}
	
	delete list_of_combinations ; // We can do this safely because, at this
		// point, because we know that it is empty.
				
	list_of_combinations = temp_list ;
	
} // sort_combinations

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Writes out the list_of_combinations.
**
** CALLING SEQUENCE:
**
**	void write_combinations()
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	list_of_combinations
**
** IMPLICIT OUTPUTS:
**
**	To output_file
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void write_combinations()
{
	combination *comb1 ;

	fprintf(output_file, "COMB_ID\tSCORE\t\tBASIC ALIGNMENTS (ID)\n\n") ;
				
	for (comb1 = (combination *)list_of_combinations->get_first_child();
		comb1 != NIL; comb1 = (combination *)
			list_of_combinations->get_next_child())
	{
		fprintf(output_file, "%s%d%s%5.2f%s",
			"C_ID",
			comb1->get_C_ID(),
			"    \t",
			comb1->get_combination_score(),
			"    \t") ;
		comb1->print_sub_alignments() ;
		fprintf(output_file, "\n") ;
	}

	fprintf(output_file, "\n") ;
} // write_combinations

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	For each combination, this function searchs for unmatched
**	portions of New and, for each such portion, creates a new 
**	basic pattern and a corresponding alignment between New 
**	and that basic pattern, and adds this new basic alignment 
**	to the combination.
**
** CALLING SEQUENCE:
**
**	void augment_set_of_combinations()
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	list_of_combinations
**
** IMPLICIT OUTPUTS:
**
**	New basic patterns and corresponding basic alignments
**	with New (stored in old_patterns and 
**	created_alignments), and new combinations (added to
**	list_of_combinations) containing the new basic alignments.
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void augment_set_of_combinations()
{
	combination *comb1 ;

	fprintf(output_file, "Start of augment_set_of_combinations()\n\n") ;

	for (comb1 = (combination *)list_of_combinations->get_first_child();
		comb1 != NIL; comb1 = (combination *)
			list_of_combinations->get_next_child())
	{
		comb1->augment_combination() ;
	}

	// Now print out augmented combinations.

	fprintf(output_file, "Augmented combinations\n\n") ;

	write_combinations() ;

} // augment_set_of_combinations

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	From a list of alternative alignments formed at the end of
**	the 'recognition' phase of processing a pattern from New,
**	this function tries to find combinations of basic alignments
**	that will fit together without overlap. It builds the best
**	of these into new composite alignments.
**
**	For the time being, the simplifying assumption is made that,
**	for every basic alignment, the sequence of hit symbols in
**	current_new_pattern is coherent, without any gaps.
**
**	In this context, 'basic alignment' simply means an alignment 
**	that has been created without being formed from the combining
**	of simpler alignments (by combine_sub_alignments()).
**
** CALLING SEQUENCE:
**
**	void combine_sub_alignments()
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	current_alignments_from_parsing
**
** IMPLICIT OUTPUTS:
**
**	A list of combinations of basic alignments and, from the best
**	of these, new composite alignments.
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void combine_sub_alignments()
{
	sequence *sub_al1 ;
	sub_alignment *sub_al2 ;
	struct sub_alignment_entry *ba_entry ;
	combination *comb1, *comb2 ;
	symbol *symbol1, *symbol_new ;
	alignment_element *al_el1 ;
	int pos_new_first, pos_new_last ;
	
	int ba_index = 0 ;
	
	list_of_combinations = new group ;

	fe_sub_alignments_array = current_alignments_from_parsing->
									get_number_of_children() ;
	sub_alignments_array =
		new sub_alignment_entry[fe_sub_alignments_array] ;

	// Convert types of alignments in current_alignments_from_parsing.

	group *temp_group = new group ;

	while (sub_al1 = (sequence *)
		current_alignments_from_parsing->extract_first_child())
	{
		sub_al2 = new sub_alignment(*sub_al1) ;
		delete sub_al1 ;
		temp_group->add_child(sub_al2) ;
	}

	while (sub_al2 = (sub_alignment *)temp_group->extract_first_child())
		current_alignments_from_parsing->add_child(sub_al2) ;

	delete temp_group ;
	
	list_for(sub_al2, sub_alignment, current_alignments_from_parsing)
	{
		if (ba_index >= fe_sub_alignments_array)
			abort_run("anomaly in sub_alignments_array[]") ;

		// Add basic alignment to sub_alignments_array[].

		ba_entry = &sub_alignments_array[ba_index++] ;
		ba_entry->sub_al1 = sub_al2 ;
					
		// Find the first and last symbol in the current pattern from
		// New that forms a hit in sub_al2. For the time being,
		// the simplifying assumption is made that, for every basic
		// alignment, the sequence of hit symbols in
		// current_new_pattern is coherent, without any gaps.
					
		pos_new_first = NULL_VALUE ;
		list_for(symbol1, symbol, sub_al2)
		{
			al_el1 = symbol1->get_al_el(0) ;
			symbol_new = (symbol *)al_el1->get_el_obj() ;
			if (symbol_new == NIL) continue ;
			else
			{
				if (pos_new_first == NULL_VALUE)
					pos_new_first = al_el1->get_orig_patt_int_pos() ;
				pos_new_last = al_el1->get_orig_patt_int_pos() ;
			}
			ba_entry->pos_new_first = pos_new_first ;
		}				
		ba_entry->pos_new_last = pos_new_last ;
	}
				
	int len_new = current_new_pattern->get_number_of_children() ;
	comb1 = new combination(len_new) ;
	combine_alignments(-1, comb1) ; // A recursive function that
		// systematically identifies combinations of basic alignments
		// that 'cover' the symbols in the current pattern from New.
				
	// Sort master list of combinations and print.
				
	sort_combinations() ;
				
	fprintf(output_file, "LEGAL COMBINATIONS OF BASIC ALIGNMENTS\n\n") ;

	write_combinations() ;

	// Remove any combination that is a subset of any other combination
	// on the list_of_combinations.
				
	group *temp_list = new group ; // This will take each combination
		// that is not a subset of any other combination.
				
	comb1 = (combination *)list_of_combinations->extract_first_child() ;
	while (comb1 != NIL)
	{
		for (comb2 = (combination *)temp_list->get_first_child();
			comb2 != NIL; comb2 = (combination *)
				temp_list->get_next_child())
		{
			if (comb1->is_subset_of(comb2)) break ;
		}
					
		if (comb2 == NIL)
		{
			// comb1 has been tested against all the combinations
			// on temp_list and has not found to be a subset of any
			// of them. This means it can be added to temp_list.
						
			temp_list->add_child(comb1) ;
		}
		else
		{
			fprintf(output_file, "Combination deleted: ") ;
			comb1->print_sub_alignments() ;
			fprintf(output_file, "%s%d%s", 
				" (C_ID",
				comb1->get_C_ID(),
				")\n") ;
			delete comb1 ;
		}
					
		comb1 = (combination *)list_of_combinations->
			extract_first_child() ;
	}
				
	delete list_of_combinations ;
	list_of_combinations = temp_list ;
				
	fprintf(output_file, "\nReduced list of combinations:\n\n") ;

	write_combinations() ;
				
	// For each combination, search for unmatched portions of New and,
	// for each such portion, create a new basic pattern and a 
	// corresponding alignment between New and that basic pattern,
	// and add this new basic alignment to the combination.

	augment_set_of_combinations() ;
				
	// Select the best (up to some limit) and convert into alignments.
				
	int i = 0 ;

	list_for(comb1, combination, list_of_combinations)	
	{
		comb1->make_composite_alignment() ;
		if (++i >= ALIGNMENTS_FOR_LEARNING)
		{
			fprintf(output_file,
				"Limit reached on list of combinations\n\n") ;
			break ;
		}
	}
				
	// Dispose of combinations and the master list of combinations.

	delete list_of_combinations ;
	delete sub_alignments_array ;
				
} // combine_sub_alignments

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	For each pattern in old_patterns, this function counts the number
**	of times it is recognised in current_alignments_from_parsing (the
**	alignments formed for ONE pattern from New). This
**	count is the maximum number of times it is 'completely' matched
**	in any ONE alignment. In this context, 'complete' matching means
**	that all the CONTENTS symbols of the given pattern are matched
**	and that all the CONTENTS symbols of all the other patterns in
**	the alignment are matched.
**
**	Notice that this method of counting avoids spurious double counting
**	that might arise from the fact that any given pattern from New
**	may appear in several different alternative alignments. Amongst
**	these alternative alignments, we are only interested in the
**	*maximum* count in any *one* alignment for a given pattern from New.
**
**	The function also performs a similar count for the symbol types
**	that appear in 'full' alignments. For the 'full' alignments in
**	current_alignments_from_parsing, the function counts the maximum
**	number of times any given symbol type appears in ONE alignment.
**
** CALLING SEQUENCE:
**
**	void count_patterns_and_symbol_types()
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	old_patterns, current_alignments_from_parsing
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void count_patterns_and_symbol_types()
{
	sequence *pattern1, *pattern2, *alignment1 ;
	int	index1, index2 ;
	symbol *symbol_type1, *symbol_type2 ;

	for (index1 = 0; index1 < size_of_old_patterns; index1++)
		patterns_counts_array[index1].max_cioa = 0 ;

	for (index1 = 0; index1 < current_alphabet_size; index1++)
		symbol_types_counts_array[index1].max_cioa = 0 ;

	// Scan through current_alignments_from_parsing looking for
	// alignments where all the CONTENTS symbols are matched and where
	// all the symbols in current_new_pattern are matched. At this stage,
	// as a result of 'weeding', all of the alignments should conform
	// to this rule. However, each one is checked to ensure that
	// this is true.

	// For each such alignment, count the number of occurrences
	// of each pattern from Old in the alignment. Wherever the count
	// for a given pattern exceeds the existing count for that pattern,
	// update the existing count with the new value.

	// Likewise, for each such alignment, count the number of occurrences
	// of each symbol type in symbols_in_old_after_learning. Wherever
	// the count for a given symbol type exceeds the existing count for
	// that symbol type, update the existing count with the new value.

	int sequence_depth, row ;

	list_for(alignment1, sequence, current_alignments_from_parsing)
	{
		if (alignment1->is_full_alignment() == false)
			abort_run("Anomaly in current_alignments_from_parsing.") ;

		// We have an alignment in which all the CONTENTS symbols
		// in Old patterns are matched as are all the CONTENTS
		// symbols in current_new_pattern. For future reference,
		// a pointer to this alignment is stored in full_alignments_from_parsing.

		full_alignments_from_parsing->add_child(alignment1) ;

		// Now, for each of the patterns in the alignment
		// (except current_new_pattern), update count_in_one_alignment
		// and max_cioa. Do likewise for the symbol types that appear
		// in the same patterns.

		for (index1 = 0; index1 < size_of_old_patterns; index1++)
			patterns_counts_array[index1].count_in_one_alignment = 0 ;

		for (index1 = 0; index1 < current_alphabet_size; index1++)
			symbol_types_counts_array[index1].count_in_one_alignment = 0 ;

		sequence_depth = alignment1->get_sequence_depth() ;
		for (row = 1; row < sequence_depth; row++)
		{
			pattern1 = alignment1->get_row_pattern(row) ;
			for (index1 = 0; index1 < size_of_old_patterns; index1++)
			{
				pattern2 = patterns_counts_array[index1].pattern1 ;
				if (pattern1 == pattern2) break ;
			}

			if (index1 >= size_of_old_patterns)
				abort_run("pattern1 not found in \
					count_patterns_and_symbol_types()") ;

			(patterns_counts_array[index1].count_in_one_alignment)++ ;
			if (patterns_counts_array[index1].count_in_one_alignment >
				patterns_counts_array[index1].max_cioa)
					patterns_counts_array[index1].max_cioa =
						patterns_counts_array[index1].count_in_one_alignment ;

			// Now count occurrences of symbol types.

			list_for(symbol_type1, symbol, pattern2)
			{
				for (index2 = 0; index2 < current_alphabet_size; index2++)
				{
					symbol_type2 =
						symbol_types_counts_array[index2].symbol_type ;
					if (symbol_type1->name_matches(symbol_type2)) break ;
				}

				if (index2 >= current_alphabet_size)
					abort_run("symbol_type1 not found in \
					count_patterns_and_symbol_types()") ;

			(symbol_types_counts_array[index2].
					count_in_one_alignment)++ ;
				if (symbol_types_counts_array[index2].count_in_one_alignment >
						symbol_types_counts_array[index2].max_cioa)
					symbol_types_counts_array[index2].max_cioa =
						symbol_types_counts_array[index2].count_in_one_alignment ;
			}
		}
	}

	// Now add max_cioa values to frequency counts for the patterns
	// in old_patterns.

	fprintf(output_file, "Frequency values for current pattern from New: ") ;
	current_new_pattern->write_tree_object(PRINT_NO_FREQUENCIES) ;

	fprintf(output_file, "Patterns:\n\n") ;

	int counter = 0 ;
	for (index1 = 0; index1 < size_of_old_patterns; index1++)
	{
		pattern1 = patterns_counts_array[index1].pattern1 ;
		patterns_counts_array[index1].frequency += patterns_counts_array[index1].max_cioa ;
		fprintf(output_file, "%s%d%s%d%s%d%s",
			"ID",
			pattern1->get_ID(),
			" = ",
			patterns_counts_array[index1].max_cioa,
			" (",
			patterns_counts_array[index1].frequency,
			")") ;

		counter++ ;
		if (counter >= OUTPUT_COUNTER_1)
		{
			if (index1 + 1 >= size_of_old_patterns)
				fprintf(output_file, ".\n\n") ;
			else fprintf(output_file, ",\n") ;
			counter = 0 ;
		}
		else if (index1 + 1 >= size_of_old_patterns)
			fprintf(output_file, ".\n\n") ;
		else fprintf(output_file, ", ") ;
	}

	// And add max_cioa values to frequency counts for the symbol types
	// in symbols_in_old_after_learning. At the same time, print out
	// values for max_cioa and cumulative frequency.

	fprintf(output_file, "Symbol types:\n\n") ;

	counter = 0 ;
	for (index2 = 0; index2 < current_alphabet_size; index2++)
	{
		symbol_type1 = symbol_types_counts_array[index2].symbol_type ;
		symbol_types_counts_array[index2].frequency +=
			symbol_types_counts_array[index2].max_cioa ;
		fprintf(output_file, "%s%s%d%s%d%s",
			symbol_type1->get_name(),
			" = ",
			symbol_types_counts_array[index2].max_cioa,
			" (",
			symbol_types_counts_array[index2].frequency,
			")") ;

		counter++ ;
		if (counter >= OUTPUT_COUNTER_1)
		{
			if (index2 + 1 >= current_alphabet_size)
				fprintf(output_file, ".\n\n") ;
			else fprintf(output_file, ",\n") ;
			counter = 0 ;
		}
		else if (index2 + 1 >= current_alphabet_size)
			fprintf(output_file, ".\n\n") ;
		else fprintf(output_file, ", ") ;
	}

} // count_patterns_and_symbol_types

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Checks whether or not an alignment has all its New symbols matched
**	and all CONTENTS symbols in its Old patterns matched.
**
** CALLING SEQUENCE:
**
**	bool sequence::is_full_alignment()
**
** FORMAL ARGUMENTS:
**
**	Return value:	true if all New symbols are matched and all CONTENTS
**					symbols in Old patterns.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

bool sequence::is_full_alignment()
{
	symbol *col1, *symbol1 ;
	int i ;

	if (all_new_CONTENTS_symbols_matched() == false) return(false) ;

	list_for(col1, symbol, this)
	{
		if (col1->is_a_hit()) continue ;

		// Find the single symbol in an Old pattern in this column.
		// There should not be any New symbol because, in general,
		// unmatched New symbols are not entered into alignments.

		for (i = 1; i < sequence_depth; i++)
		{
			symbol1 = col1->get_row_symbol(i) ;
			if (symbol1 != NIL) break ;
		}

		if (symbol1 == NIL)
			abort_run("Empty column in sequence::is_full_alignment()") ;

		if (symbol1->get_status() == CONTENTS)
			return(false) ;
	}

	return(true) ;

} // sequence::is_full_alignment

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	From the list of alignments in current_alignments_from_parsing, this
**	function weeds out the ones containing unmatched CONTENTS symbols. It
**	also weeds out alignments in which one or more symbols from
**	current_new_pattern are not matched.
**
**	The reason that any have survived to this stage is that, in the
**	recognise() function, it is necessary to allow alignments which
**	contain unmatched symbols because otherwise the function
**	may not be able to build up full alignments.
**
** CALLING SEQUENCE:
**
**	void weed_out_partial_alignments(group *cafp)
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
**	cafp:			Normally, current_alignments_from_parsing.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	Weeded cafp
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void weed_out_partial_alignments(group *cafp)
{
	int max_size = cafp->get_number_of_children(), index1 = 0,
		index_limit ;
	sequence **alignments_to_be_weeded, *al1 ;

	alignments_to_be_weeded = new sequence*[max_size] ;

	list_for(al1, sequence, cafp)
	{
		if (al1->is_full_alignment() == true) continue ;
		alignments_to_be_weeded[index1++] = al1 ;
	}

	index_limit = index1 ;

	// Now weed out the alignments identified in alignments_to_be_weeded[].

	for (index1 = 0; index1 < index_limit; index1++)
	{
		al1 = alignments_to_be_weeded[index1] ;
		cafp->extract_child(al1) ;
		delete al1 ;		
	}

	delete[] alignments_to_be_weeded ;
} // weed_out_partial_alignments

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Compress the patterns in New in terms of patterns in Old (if any).
**
** CALLING SEQUENCE:
**
**	void compress_new_into_old(bool parsing_and_learning)
**
** FORMAL ARGUMENTS:
**
**	Return value:			void
**
**	parsing_and_learning:	If true, do both parsing (including
**							the formation of partial matches of
**							patterns in Old) and learning from those
**							partial matches. Otherwise, do 
**							only 'exact' parsing where all and only
**							the CONTENTS symbols of patterns are
**							recognised.
**
** IMPLICIT INPUTS:
**
**	The corpus
**
** IMPLICIT OUTPUTS:
**
**	Alignments and new patterns
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void compress_new_into_old(bool parsing_and_learning)
{
	sequence *best_alignment, *al1 ;
	int counter ;
	list_element *pos2 ;

	total_number_of_alignments = 0 ;
	new_pattern_number = 0 ;

	list_for(current_new_pattern, sequence, new_patterns)
	{
		// Outer loop: taking successive patterns from New.
			
		new_pattern_number++ ;
			
		fprintf(output_file, "NEXT PATTERN FROM NEW:\n\n") ;
		current_new_pattern->print_ID() ;
		fprintf(output_file, " ") ;
		current_new_pattern->write_tree_object(PRINT_NO_FREQUENCIES) ;

		current_new_pattern->write_with_details() ;

		if (parsing_and_learning)
		{
			// Prepare a "receptacle_pattern".
			
			receptacle_pattern = new sequence ; // Receives 
				// symbols from current_new_pattern, one by one, after 
				// each one has been matched against symbols in Old 
				// (including any symbols in receptacle_pattern).
			old_patterns->add_child(receptacle_pattern) ;
			naive_grammar->add_pattern(receptacle_pattern) ;
		}

		window = 0 ;
		
		// Get windows in succession until current_new_pattern
		// is exhausted. The process of measuring off successive
		// windows occurs in find_best_matches().
			
		cns_el_pos = NIL ; // This will be used to step through
			// current_new_pattern.
		current_new_pattern->initialise() ;
		while (true)
		{
			// Inner loop: the next window from 
			// current_new_pattern will be obtained in 
			// find_best_matches within recognise() and
			// compressed as far as possible in terms of
			// the patterns in old_patterns and in alignments
			// that have been found.
				
			cycle = 0 ; // This counter counts the number of
				// cycles within the 'recognise()' function
				// (below) are made for this new window.
				
			window++ ; // Count the number of windows in
				// each pattern from New.
				
			best_score_this_window = fail_score ;

			// Diagnostic

			if (driving_patterns_and_alignments->is_empty() == false)
				abort_run("driving_patterns_and_alignments not empty.") ;

			// Do the current window of current_new_pattern.
		
			fprintf(output_file, "DRIVING PATTERN FROM NEW (") ;
			print_pattern_window_cycle() ;
			fprintf(output_file, ")\n\n") ;
			current_new_pattern->print_ID() ;
			fprintf(output_file, " ") ;
			current_new_pattern->write_tree_object(PRINT_NO_FREQUENCIES) ;

			driving_patterns_and_alignments->add_child(current_new_pattern) ;

			while (recognise(parsing_and_learning)) ;

			if (current_alignments_from_parsing->is_empty())
			{
				fprintf(output_file, 
					"BEST ALIGNMENT FOUND SINCE START: None\n\n") ;
					
				fprintf(output_file, "END OF ") ;
				print_pattern_window_cycle() ;
				fprintf(output_file, "\n\n") ;

				if (cns_el_pos == NIL)
					// End of current_new_pattern
					// has been reached.
					break ;
				else continue ; 
			}
				
			current_alignments_from_parsing->
				sort_by_compression_difference() ;

			best_alignment = (sequence *)
				current_alignments_from_parsing->get_first_child() ;

			if (parsing_and_learning)
			{
				fprintf(output_file, "%s%d%s",
				"BEST ALIGNMENTS FOUND UP TO ",
				BEST_ALIGNMENT_PRINT_LIMIT,
				" (") ;
				print_pattern_window_cycle() ;
				fprintf(output_file, ")\n\n") ;

				counter = 0 ;
				list_for(al1, sequence, current_alignments_from_parsing)
				{
					counter++ ;
					al1->write_out_fully(NIL, NIL, NIL,
						write_section_chars_length, NULL_VALUE, false) ;
					if (counter >= BEST_ALIGNMENT_PRINT_LIMIT) break ;
				}
			}
			else
			{
				fprintf(output_file, 
					"BEST ALIGNMENTS (FULL) BEFORE WEEDING\n\n") ;

				counter = 0 ;
				list_for(al1, sequence, current_alignments_from_parsing)
				{
					counter++ ;
					al1->write_out_fully(NIL, NIL, NIL,
						write_section_chars_length, NULL_VALUE, false) ;
					// if (counter >= BEST_ALIGNMENT_PRINT_LIMIT) break ;
				}
				
				// Weed out alignments containing unmatched
				// CONTENTS symbols and/or unmatched IDENTIFICATION
				// symbols in New.

				weed_out_partial_alignments(current_alignments_from_parsing) ;

				if (current_alignments_from_parsing->is_empty())
					fprintf(output_file, "No alignments left after weeding.\n\n") ;
				else
				{
					fprintf(output_file, "BEST ALIGNMENTS (FULL) AFTER WEEDING:\n\n") ;

					list_for(al1, sequence, current_alignments_from_parsing)
					{
						al1->write_out_fully(NIL, NIL, NIL,
							write_section_chars_length, NULL_VALUE, false) ;
					}
				}
			}

			// Look for matches for the next window from the
			// current pattern from New. Search for good 
			// combinations of basic alignments and create 
			// new alignments from them.
				
			// combine_sub_alignments() ;

			// Examine best_alignment for unmatched portions of New
			// In those cases, create new patterns and classes 
			// (where appropriate) and a new alignment that 
			// incorporates these patterns and classes.

			fprintf(output_file, "END OF ") ;
			print_pattern_window_cycle() ;
			fprintf(output_file, "\n\n") ;
				
			// Test for the end of current_new_pattern.
				
			if (cns_el_pos == NIL) break ;
				
		} // Inner loop: get next window from current_new_pattern.
			
		if (parsing_and_learning)
		{
			// Print out receptacle pattern now that it has been
			// completed.

			fprintf(output_file, "%s%d%s%d%s",
				"From current_new_pattern ID",
				current_new_pattern->get_ID(),
				",\nthe corresponding pattern stored in Old is ID",
				receptacle_pattern->get_ID(),
				": ") ;

			receptacle_pattern->write_tree_object(PRINT_NO_FREQUENCIES)	;

			receptacle_pattern->write_with_details() ;

			// Extract patterns and classes.

			current_alignments_from_parsing->
				sort_by_compression_difference() ;

			counter = 0 ;
			pos2 = NIL ;
			while (al1 = (sequence *)current_alignments_from_parsing->
				get_next_child_by_el_pos(&pos2))
			{
				extract_patterns_and_classes(al1) ;
				counter++ ;
				if (counter >= extraction_limit) break ;
			}
		}
		else count_patterns_and_symbol_types() ;

		// Count the number of times each pattern in Old has
		// been recognised and, for each pattern, add its
		// count to the cumulative frequency of the pattern.

		// Move alignments from current_alignments_from_parsing
		// to global_alignments_from_parsing.

		while (al1 = (sequence *)current_alignments_from_parsing->
			extract_first_child())
				global_alignments_from_parsing->add_child(al1) ;

		/* Calculate probabilities of inferences from the best
		alignment and all other alignments which encode
		the same symbols from New. */
	
		if (probabilities) 
			probabilities_of_inferences(best_alignment) ;

	} // Outer loop: taking successive patterns from New. Get next 
		// current_new_pattern from New.
} // compress_new_into_old

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	For the 'full' alignments stored in 
**	full_alignments_from_parsing, re-compute codes and 
**	encoding costs. Compute total_cost for each of the patterns in
**	new_patterns and old_patterns.
**
** CALLING SEQUENCE:
**
**	void compute_scores()
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	full_alignments_from_parsing
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void compute_scores()
{
	sequence *al1, *new_code, *pattern1 ;
	group *selected_old_patterns = new group ;

	created_code_patterns->delete_children() ;

	for (al1 = (sequence *)full_alignments_from_parsing->
		get_first_child();al1 != NIL ; al1 = (sequence *)
			full_alignments_from_parsing->get_next_child())
	{
		al1->correct_column_values() ; // Each column receives the
			// frequency and bit_cost values for the symbol type
			// within the column (changed as a result of the
			// application of find_symbol_type_information_costs()).
		new_code = al1->make_code(true) ;
		created_code_patterns->add_child(new_code) ;
	}

	// Sort full_alignments_from_parsing and print results.

	fprintf(output_file, "SORTED FULL ALIGNMENTS FROM PARSING:\n\n") ;

	full_alignments_from_parsing->sort_children() ;

	list_for(al1, sequence, full_alignments_from_parsing)
	{
		fprintf(output_file, "%s%d%s%1.2f%s%1.2f%s%1.2f%s",
			"ID",
			al1->get_ID(),
			", NSC = ",
			al1->get_new_symbols_cost(),
			", EC = ",
			al1->get_encoding_cost(),
			", CD = ",
			al1->get_compression_difference(),
			": ") ;
		al1->write_tree_object(PRINT_NO_FREQUENCIES) ;
	}

	// Compute scores for patterns in new_patterns

	list_for(pattern1, sequence, new_patterns)
		pattern1->compute_costs() ;

	// Compute scores for patterns in old_patterns.

	list_for(pattern1, sequence, old_patterns)
		pattern1->compute_costs() ;

	// Compile a list of the patterns in the alignments in
	// full_alignments_from_parsing, without duplicates.

	int alignment_depth, row ;
	list_for(al1, sequence, full_alignments_from_parsing)
	{
		alignment_depth = al1->get_sequence_depth() ;
		for (row = 1; row < alignment_depth; row++)
		{
			pattern1 = al1->get_row_pattern(row) ;

			selected_old_patterns->
				add_child_without_duplicates(pattern1) ;
		}
	}

	// Print out patterns in selected_old_patterns with frequencies
	// and scores.

	fprintf(output_file, "PATTERNS IN 'FULL' ALIGMENTS:\n\n") ;

	selected_old_patterns->sort_by_ID() ;
	for (pattern1 = (sequence *)selected_old_patterns->
		get_first_child();pattern1 != NIL ; pattern1 = (sequence *)
			selected_old_patterns->get_next_child())
	{
		fprintf(output_file, "%s%d%s%d%s%3.2f%s%3.2f%s",
			"ID",
			pattern1->get_ID(),
			", Fr = ",
			pattern1->get_frequency(),
			", EC = ",
			pattern1->get_encoding_cost(),
			", TC = ",
			pattern1->get_total_cost(), 
			": ") ;
		pattern1->write_tree_object(PRINT_NO_FREQUENCIES) ;
	}

	selected_old_patterns->release_children() ;
	delete selected_old_patterns ;

} // compute_scores

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Given a list of 'full' alignments from the parsing of patterns in New
**	(in full_alignments_from_parsing), this function compiles
**	a list of alternative grammars for the patterns in New and, for
**	each one, calculates a value for G (the size of the grammar) 
**	and E (the size of New when it has been encoded in terms of 
**	the grammar).
**
**	It is assumed that full_alignments_from_parsing have been sorted
**	by CD score. Provided this function is called after 
**	compute_scores() (which it currently is, in 
**	sifting_and_sorting()), this assumption should be correct.
**
**	This function uses two constraints: full_alignment_limit (which
**	restricts the number of alternative alignments that may be
**	considered for any one pattern from New; and grammar_limit (which
**	restricts the number of alternative grammars that may be developed
**	at any one time.
**
**	This function also measures the sizes of the best grammars found
**	for successive patterns from New and outputs this information in
**	a form that is suitable for plotting.
**
** CALLING SEQUENCE:
**
**	void compile_alternative_grammars()
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	full_alignments_from_parsing
**
** IMPLICIT OUTPUTS:
**
**	set_of_grammars
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void compile_alternative_grammars()
{
	sequence *pattern_new, *coded_pattern_new, *pattern_row_0, 
		*full_alignment ;
	int alignment_counter, grammar_counter, new_pattern_counter = 0 ;
	group *temp_set_of_grammars = new group ;
	grammar *grammar1, *grammar2, *best_grammar ;
	double lowest_score, temp_score, score_for_best_grammar,
		G_for_best_grammar, E_for_best_grammar, G_naive = 0,
		E_naive = 0, GE_naive, G_raw = 0 ;

	naive_grammar->initialise() ;

	fprintf(plot_file, "%s%s%s%s%s%s%s%s%s",
		"new pattern, ",
		"G, ",
		"E, ",
		"G + E, ",
		"G (naive), ",
		"E (naive), ",
		"G + E (naive), ",
		"G (raw)",
		"\n") ;

	list_for(pattern_new, sequence, new_patterns)
	{
		coded_pattern_new = naive_grammar->get_next_pattern() ;
		alignment_counter = 0 ;
		new_pattern_counter++ ;
		list_for(full_alignment, sequence, full_alignments_from_parsing)
		{
			// Select alignments that contain pattern_new in row 0.

			pattern_row_0 = full_alignment->get_row_pattern(0) ;
			if (pattern_row_0 != pattern_new) continue ;

			alignment_counter++ ;

			// full_alignment has pattern_new in its row 0.

			if (set_of_grammars->is_empty())
			{
				grammar1 = new grammar ;
				temp_set_of_grammars->add_child(grammar1) ;
				grammar1->compile_grammar(full_alignment) ;
				grammar1->compute_G() ;
				grammar1->increment_E(full_alignment->
					get_encoding_cost()) ;
			}
			else
			{
				// Step through the list of grammars as it was
				// at the end of processing the previous pattern
				// from New.

				list_for(grammar1, grammar, set_of_grammars)
				{
					grammar2 = new grammar(*grammar1) ;
					temp_set_of_grammars->add_child(grammar2) ;
					grammar2->compile_grammar(full_alignment) ;
					grammar2->compute_G() ;
					grammar2->increment_E(full_alignment->
						get_encoding_cost()) ;
				}
			}

			if (alignment_counter >= full_alignment_limit) break ; // This cut-off
				// assumes that full alignments in 
				// full_alignments_from_parsing have been sorted in
				// descending order of CD score.
		}

		// Delete all the grammars in set_of_grammars

		while (grammar1 = (grammar *)
			set_of_grammars->extract_first_child())
				delete grammar1 ;

		// Compute the score of each grammar in temp_set_of_grammars
		// as G + E. At the same time, count the number of grammars
		// that have been formed.

		grammar_counter = 0 ;
		list_for(grammar1, grammar, temp_set_of_grammars)
		{
			grammar_counter++ ;
			grammar1->sum_G_and_E() ;
		}

		fprintf(output_file, "%s%d%s%d%s",
			"Number of grammars formed for New pattern ID",
			pattern_new->get_ID(),
			" = ",
			grammar_counter,
			"\n\n") ;

		// Now transfer newly-updated grammars from temp_set_of_grammars
		// to set_of_grammars in ascending order of their scores. When
		// the number of grammars has reached grammar_limit, stop
		// transfering grammars and delete the remaining grammars
		// instead.

		grammar_counter = 0 ;
		while (temp_set_of_grammars->is_empty() == false)
		{
			grammar_counter++ ;
			if (grammar_counter > grammar_limit)
			{
				while (grammar1 = (grammar *)
					temp_set_of_grammars->extract_first_child())
						delete grammar1 ;

				fprintf(output_file, "%s%d%s",
					"Grammar limit reached for New pattern, ID",
					pattern_new->get_ID(),
					"\n\n") ;

				break ;
			}
			else
			{
				lowest_score = HIGH_VALUE ;
				list_for(grammar1, grammar, temp_set_of_grammars)
				{
					temp_score = grammar1->get_score() ;
					if (temp_score < lowest_score)
					{
						lowest_score = temp_score ;
						best_grammar = grammar1 ;
					}
				}
				temp_set_of_grammars->extract_child(best_grammar) ;
				set_of_grammars->add_child(best_grammar) ;
			}
		}

		if (grammar_counter > 0)
		{
			best_grammar = (grammar *)set_of_grammars->get_first_child() ;
			G_for_best_grammar = best_grammar->get_G() ;
			E_for_best_grammar = best_grammar->get_E() ;
			score_for_best_grammar = best_grammar->get_score() ;
			G_naive += coded_pattern_new->get_total_cost() ;
			G_raw += pattern_new->get_total_cost() ;
			E_naive += coded_pattern_new->get_encoding_cost() ;
			GE_naive = G_naive + E_naive ;
			naive_grammar->increment_E(E_naive) ;
			naive_grammar->increment_G(G_naive) ;
			naive_grammar->sum_G_and_E() ;

			fprintf(plot_file, 
				"%d%s%3.2f%s%3.2f%s%3.2f%s%3.2f%s%3.2f%s%3.2f%s%3.2f%s",
				new_pattern_counter,
				", ",
				G_for_best_grammar,
				", ",
				E_for_best_grammar,
				", ",
				score_for_best_grammar,
				", ",
				G_naive,
				", ",
				E_naive,
				", ",
				GE_naive,
				", ",
				G_raw,
				"\n") ;
		}
		else fprintf(output_file, "Grammar counter == 0.\n\n") ;
	}

	// Now write out the grammars in set_of_grammars.

	fprintf(output_file, "FINAL SET OF GRAMMARS:\n\n") ;
	list_for(grammar1, grammar, set_of_grammars)
	{
		grammar1->sort_by_ID() ;
		grammar1->write_grammar() ;
	}

	delete temp_set_of_grammars ;

} // compile_alternative_grammars

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Re-parses the patterns in New, looking for alignments in which
**	all the CONTENTS symbols are matched and, from these, counting
**	frequencies of patterns and symbols. The function ends by computing
**	a set of alternative grammars.
**
** CALLING SEQUENCE:
**
**	void sifting_and_sorting()
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	patterns in New
**
** IMPLICIT OUTPUTS:
**
**	Sets of alternative grammars
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void sifting_and_sorting()
{
	sequence *pattern1 ;

	// Diagnostic

	if (driving_patterns_and_alignments->is_empty() == false)
		abort_run("Non-empty driving_patterns_and_alignments") ;

	write_message("RE-PARSING OF PATTERNS IN NEW") ;
	
	size_of_old_patterns = old_patterns->get_number_of_children() ;
	patterns_counts_array =
		new patterns_counts_record[size_of_old_patterns] ;

	int index1 = 0 ;

	list_for(pattern1, sequence, old_patterns)
	{
		patterns_counts_array[index1].pattern1 = pattern1 ;
		patterns_counts_array[index1].frequency = 0 ;
		index1++ ;
	}

	symbol_types_counts_array =
		new symbol_types_counts_record[current_alphabet_size] ;

	index1 = 0 ;
	symbol *symbol_type ;

	list_for(symbol_type, symbol, symbols_in_old_after_learning)
	{
		symbol_types_counts_array[index1].symbol_type = symbol_type ;
		symbol_types_counts_array[index1].frequency = 0 ;
		index1++ ;
	}

	compress_new_into_old(false) ;

	// Print out a list of the full_alignments_from_parsing.

	fprintf(output_file, "CURRENT FULL ALIGNMENTS FROM PARSING:\n\n") ;

	sequence *al1 ;

	list_for(al1, sequence, full_alignments_from_parsing)
	{
		fprintf(output_file, "%s%d%s",
			"ID",
			al1->get_ID(),
			": ") ;

		al1->write_tree_object(PRINT_NO_FREQUENCIES) ;
	}

	// Transfer final frequency values to patterns in Old and print
	// them out at the same time.

	fprintf(output_file, "FREQUENCY VALUES FOR PATTERNS IN OLD:\n\n") ;

	int frequency ;
	for (index1 = 0; index1 < size_of_old_patterns; index1++)
	{
		pattern1 = patterns_counts_array[index1].pattern1 ;
		frequency = patterns_counts_array[index1].frequency ;
		pattern1->set_frequency(frequency) ;
		fprintf(output_file, "%s%d%s%d%s",
			"ID",
			pattern1->get_ID(),
			", Fr = ",
			frequency,
			" ") ;
		pattern1->write_tree_object(PRINT_NO_FREQUENCIES) ;
	}
	
	// Transfer final frequency values to symbol types in
	// symbols_in_old_after_learning and print them out at
	// the same time.

	int counter1 = 0, counter2 = 0 ;
	fprintf(output_file, "FREQUENCY VALUES FOR SYMBOL TYPES IN OLD:\n\n") ;

	for (index1 = 0; index1 < current_alphabet_size; index1++)
	{
		symbol_type = symbol_types_counts_array[index1].symbol_type ;
		frequency = symbol_types_counts_array[index1].frequency ;
		symbol_type->set_frequency(frequency) ;
		fprintf(output_file, "%s%s%d",
			symbol_type->get_name(),
			": ",
			frequency) ;

		counter1++ ;
		counter2++ ;

		if (counter2 == current_alphabet_size)
			fprintf(output_file, ".\n\n") ;
		else if (counter1 == OUTPUT_COUNTER_2)
		{
			fprintf(output_file, ",\n\n") ;
			counter1 = 0 ;
		}
		else fprintf(output_file, ", ") ;
	}
	
	delete[] patterns_counts_array ;
	delete[] symbol_types_counts_array ;

	// Compute bit_cost for each symbol type. Assign frequencies
	// and costs to instances of the symbols in new_patterns
	// and old_patterns.

	find_symbol_type_information_costs(symbols_in_old_after_learning) ; 

	compute_scores() ; // For the 'full' alignments stored in 
		// full_alignments_from_parsing, re-compute codes and 
		// encoding costs. Compute total_cost for each of the patterns in
		// new_patterns and old_patterns. This function also corrects
		// the values for frequency and bit_cost for each column
		// in each alignment in full_alignments_from_parsing.

	compile_alternative_grammars() ;

} // sifting_and_sorting

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Assigns status and types to symbols in corpus, compiles an alphabet
**	of symbols in the corpus and prints out its size, compiles a list
**	of classes, finds frequencies of symbols, computes corresponding
**	information costs and assigns these values to symbols in the
**	corpus.
**
** CALLING SEQUENCE:
**
**	void process_symbols()
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	The corpus, original_alphabet_size
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void process_symbols()
{
	if (!identification_symbols_marked)
	{	if (phase == 1) assign_status_to_symbols() ;
	}
	else
	{
		// Make sure that left and right brackets at the beginnings
		// and ends of patterns are marked as having
		// IDENTIFICATION status.

		sequence *pattern1 ;
		symbol *symbol1 ;

		list_for(pattern1, sequence, new_patterns)
		{
			symbol1 = (symbol *)pattern1->get_first_child() ;
			if (*(symbol1->get_name()) == '<')
				symbol1->set_status(IDENTIFICATION) ;
			symbol1 = (symbol *)pattern1->get_last_child() ;
			if (*(symbol1->get_name()) == '>')
				symbol1->set_status(IDENTIFICATION) ;
		}

		list_for(pattern1, sequence, old_patterns)
		{
			symbol1 = (symbol *)pattern1->get_first_child() ;
			if (*(symbol1->get_name()) == '<')
				symbol1->set_status(IDENTIFICATION) ;
			symbol1 = (symbol *)pattern1->get_last_child() ;
			if (*(symbol1->get_name()) == '>')
				symbol1->set_status(IDENTIFICATION) ;
		}
	}

	assign_type_to_symbols() ;

	original_symbols_in_corpus = 
		corpus->compile_alphabet(&original_alphabet_size) ; 
		// This function needs to follow the previous one to ensure 
		// that type and status information is on each symbol in 
		// the alphabet.

	fprintf(output_file, "%s%d%s",
		"Original alphabet size = ",
		original_alphabet_size,
		"\n\n") ;

	find_symbol_frequencies(original_symbols_in_corpus, 
		true, true) ; /* Find the frequency of
		each symbol type by scanning over the patterns in the corpus
		and adding the frequency of the pattern to the frequency of the
		symbol type for each occurrence of each symbol in the
		pattern in which it occurs. */
	
	find_symbol_type_information_costs(original_symbols_in_corpus) ; 
		/* From the frequencies of symbol types this
		function computes an information cost for each symbol type 
		in the corpus. Calculates average_symbol_type_cost. 
		In addition, the function assigns the frequency of its type 
		to each symbol in the corpus and the associated cost, 
		calculated by the Shannon-Fano-Elias method. */
		
} // process_symbols

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Applies create_patterns_and_sort() iteratively until no further
**	compression is achieved.
**
** CALLING SEQUENCE:
**
**	void repeat_cpas()
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

void repeat_cpas()
{
	grammar *grammar1, *grammar2 ;
	group *list_of_patterns ;
	sequence *pattern1, *pattern2 ;

	// group *list_of_patterns, *copy_list_of_patterns ;

	phase = 1 ;

	create_patterns_and_sort() ;

	grammar1 = (grammar *)set_of_grammars->get_first_child() ;
	if (grammar1 == NIL)
		fprintf(output_file, "No grammar produced in repeat_cpas().\n\n") ;
	else
	{
		fprintf(output_file, "NAIVE GRAMMAR:\n\n") ;

		naive_grammar->write_grammar() ;

		fprintf(output_file, "CLEANING UP OF GRAMMARS:\n\n") ;

		// Owing to the deletion of symbols in
		// grammar::remove_unused_context_code_symbols(),
		// it is necessary to make sure that there are no alignments
		// present that contains pointers to those symbols.

		naive_grammar->release_patterns() ;
		full_alignments_from_parsing->release_children() ;
		driving_patterns_and_alignments->release_children() ;
		keep_alignments->release_children() ;

		created_alignments->delete_children() ;
		current_alignments_from_parsing->delete_children() ;
		global_alignments_from_parsing->delete_children() ;

		fprintf(output_file,
			"GRAMMARS AFTER REMOVAL OF UNUSED CONTEXT CODE SYMBOLS:\n\n") ;

		list_for(grammar1, grammar, set_of_grammars)
		{
			grammar2 = new grammar(*grammar1) ;
			set_of_cleaned_up_grammars->add_child(grammar2) ;
			list_of_patterns = grammar2->get_list_of_patterns() ;
			list_of_patterns->release_children() ;
			grammar1->initialise() ;
			while (pattern1 = grammar1->get_next_pattern())
			{
				pattern2 = new sequence(*pattern1) ;
				grammar2->add_pattern(pattern2) ;
				copies_of_grammar_patterns->add_child(pattern2) ;
			}
			grammar2->remove_unused_context_code_symbols() ;
		}

		fprintf(output_file,
			"GRAMMARS WITH NEW NAMES FOR CONTEXT CODE SYMBOLS:\n\n") ;

		list_for(grammar1, grammar, set_of_cleaned_up_grammars)
			grammar1->tidy_up_code_symbols() ;
	}

	/* Now write out statistics */
	
	fprintf(output_file, "STATISTICS\n\n") ;
	fprintf(output_file, "%s%d%s%d%s%d%s%d%s",
		"Original size of alphabet = ", original_alphabet_size,
		"\n\nCurrent size of alphabet = ", current_alphabet_size,
		"\n\nNumber of original patterns in Old = ",
		number_of_patterns_in_old,
		"\n\nFinal number of alignments = ",
		total_number_of_alignments,
		"\n\n") ;

	phase++ ;

	// list_of_patterns = best_grammar->get_list_of_patterns() ;
	// copy_list_of_patterns = new group(*list_of_patterns) ;

	// purge_lists() ; // This deletes the children of new_patterns.

	// copy_list_of_patterns->transfer_children(new_patterns) ;
	// delete copy_list_of_patterns ;
	// new_patterns->invert_symbol_status() ;

	// fprintf(output_file, "SECOND PHASE\n\n") ;

	// create_patterns_and_sort() ;
} // repeat_cpas

/*********************************************************************/

/*
** FUNCTIONAL DESCRIPTION:
**
**	Creates new patterns by parsing. This function includes an initial parsing
**	of New patterns against themselves, with the creation of new patterns
**	where full parsing is not possible, followed by a re-parsing with
**	sifting and sorting of patterns to create one or more grammars,
**	each one scored for G + E.
**
** CALLING SEQUENCE:
**
**	bool create_patterns_and_sort()
**
** FORMAL ARGUMENTS:
**
**	Return value:	true if compression is achieved, false otherwise.
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	NONE
**
** SIDE EFFECTS:
**
**	NONE
**
*/

bool create_patterns_and_sort()
{
	// Preliminaries.

	// Set up arrays and clear them.

	sort_array = new int[hit_structure_rows] ;
	leaf_array = new hit_node*[hit_structure_rows] ;
	half_list = new hit_node*[half_hit_structure_rows] ;
	alignments_array = new sequence*[hit_structure_rows] ;
	
	int index, row1, col1 ;

	for (index = 0; index < hit_structure_rows; index++)
		sort_array[index] = NULL_VALUE ;

	for (index = 0; index < hit_structure_rows; index++)
		leaf_array[index] = NIL ;

	for (index = 0; index < half_hit_structure_rows;index++)
		half_list[index] = NIL ;

	for (index = 0; index < hit_structure_rows; index++)
		alignments_array[index] = NIL ;

	// Clear hit_node_array[].
	
	for (index = 0; index < HIT_NODE_ARRAY_SIZE; index++)
	{
		hit_node_array[index].new_hit_node = NIL ;
		hit_node_array[index].hn_parent = NIL ;
	}
	
	// Clear symbol_selection_array[][].
	
	for (row1 = 0; row1 < MAX_NEW_ARRAY_ROWS; row1++)
	{
		for (col1 = 0; col1 < MCOLS; col1++)
			symbol_selection_array[row1][col1] = NIL ;
	}

	// Make counts.

	number_of_symbols_in_new =
		new_patterns->count_number_of_symbols() ; // Count the
			// number of symbols in New.

	number_of_symbols_in_old =
		old_patterns->count_number_of_symbols() ; // Count the
	// number of symbols in Old.

	// Find the numbers of patterns in New and Old.
	
	sequence *sequence1 ;

	number_of_patterns_in_new = 0 ;

	list_for(sequence1, sequence, new_patterns)
		number_of_patterns_in_new++ ;
	
	number_of_patterns_in_old = 0 ;

	list_for(sequence1, sequence, old_patterns)
		number_of_patterns_in_old++ ;
	
	// Print out the corpus.
	
	fprintf(output_file,
		"THE WHOLE CORPUS (INCL. FREQUENCIES GREATER THAN ONE)\n\n") ;
	
	if (new_patterns->is_empty())
		fprintf(output_file, "NEW PATTERNS: None\n\n") ;
	else 
	{
		fprintf(output_file, "NEW PATTERNS:\n\n") ;
		list_for(sequence1, sequence, new_patterns)
		{
			sequence1->print_ID() ;
			fprintf(output_file, " ") ;
			sequence1->write_tree_object(PRINT_ALL_FREQUENCIES) ;
		}
	}

	if (old_patterns->is_empty())
		fprintf(output_file, "OLD PATTERNS: None\n\n") ;
	else 
	{
		fprintf(output_file, "OLD PATTERNS:\n\n") ;
		list_for(sequence1, sequence, old_patterns)
		{
			sequence1->print_ID() ;
			fprintf(output_file, " ") ;
			sequence1->write_tree_object(PRINT_ALL_FREQUENCIES) ;
		}
	}

	process_symbols() ;

	// Compile classes in New and Old (if any).

	compile_class_records() ; // From the corpus, this 
		// function compiles a list of classes, without duplicates.
	
	// Compute the size of the corpus in bits.

	double size_of_corpus_at_start = corpus->total_of_symbol_sizes() ;

	fprintf(output_file, "%s%1.2g%s",
		"Size of corpus at start: ",
		size_of_corpus_at_start,
		"\n\n") ;
	
	// Now find EC for patterns in Old.
		
	list_for(sequence1, sequence, old_patterns)
		sequence1->compute_costs() ;
		
	// Notice that the following block must come after the function
	// assign_status_to_symbols() (called in process_symbols(), above)
	// because the former depends on the original order of the symbols.
		
	compute_scaling_factors() ; /* Computes the factors to be applied
		to reduce the effective info cost of a hit symbol, depending
		on the gap between the given hit and the previous hit. */
		
	write_new_and_old() ;
		
	/* Compress the corpus. */

	compress_new_into_old(true) ;

	// Remove references to alignments within each class record in
	// set_of_class_records.

	class_record *class1 ;
	list_for(class1, class_record, set_of_class_records)
		class1->release_alignments() ;

	// Write out current_alignments_from_parsing and then purge them.

	current_alignments_from_parsing->
		write_patterns("Alignments from parsing:", NULL_VALUE) ;
	
	sequence *alignment ;
	while(true)
	{
		alignment = (sequence *)current_alignments_from_parsing->
			extract_first_child() ;
		if (alignment == NIL) break ;
		delete alignment ;
	}

	fprintf(output_file, "CREATED PATTERNS:\n\n") ;

	old_patterns->sort_by_ID() ;

	old_patterns->write_patterns("Created basic patterns:",
		BASIC_PATTERN) ;
	
	old_patterns->write_patterns("Created abstract patterns:",
		ABSTRACT_PATTERN) ;

	// Print out old_patterns.

	old_patterns->write_patterns("PATTERNS IN OLD (AUGMENTED):",
		NULL_VALUE) ;

	created_alignments->sort_by_ID() ;
	created_alignments->write_patterns("Created sub-alignments:",
		SUB_ALIGNMENT) ;
	
	created_alignments->write_patterns("Created composite alignments:",
		COMPOSITE_ALIGNMENT) ;

	created_code_patterns->sort_by_ID() ;
	created_code_patterns->write_patterns("Created codes:",
		NULL_VALUE) ;
	
	// Compile a list of symbols in old_patterns and print them out.

	symbols_in_old_after_learning = 
		old_patterns->compile_alphabet(&current_alphabet_size) ;

	fprintf(output_file, "Symbols in Old after learning:\n\n") ;

	symbol *symbol_type ;

	int counter1 = 0, counter2 = 0 ;
	list_for(symbol_type, symbol, symbols_in_old_after_learning)
	{
		fprintf(output_file, "%s",
			symbol_type->get_name()) ;

		counter1++ ;
		counter2++ ;

		if (counter2 == current_alphabet_size)
			fprintf(output_file, ".\n\n") ;
		else if (counter1 == OUTPUT_COUNTER_2)
		{
			fprintf(output_file, ",\n\n") ;
			counter1 = 0 ;
		}
		else fprintf(output_file, ", ") ;
	}

	fprintf(output_file, "%s%d%s",
		"Current alphabet size = ",
		current_alphabet_size,
		"\n\n") ;

	// Sort the class_records and print them out.

	sort_and_print_class_records() ;

	// Re-parse New, count frequencies of patterns and symbols,
	// re-compute codes and encoding costs for 'full' alignments,
	// select the best alignment for each pattern from New,
	// and find the patterns from Old that are used in those
	// alignments. This set of patterns should represent the best
	// grammar for the patterns from New which the system has found.

	sifting_and_sorting() ;

	grammar *best_grammar = (grammar *)set_of_grammars->get_first_child() ;

	double size_of_best_grammar ;
	if (best_grammar != NIL)
	{
		group *list_of_patterns = best_grammar->get_list_of_patterns() ;
		size_of_best_grammar =
			list_of_patterns->total_of_symbol_sizes() ;

		fprintf(output_file, "%s%1.2g%s",
			"Size of best grammar found: ",
			size_of_best_grammar,
			"\n\n") ;
	}
	else fprintf(output_file, "No best grammar found.\n\n") ;
	
	delete sort_array ;
	delete leaf_array ;
	delete half_list ;
	delete alignments_array ;

	if (size_of_corpus_at_start > size_of_best_grammar)
		return(true) ;
	else return(false) ;

} // create_patterns_and_sort

/*********************************************************************/

/* The main part of SP */

/*
** FUNCTIONAL DESCRIPTION:
**
**	Entry point to the SP system.
**
** CALLING SEQUENCE:
**
**	void SP_main()	  
**
** FORMAL ARGUMENTS:
**
**	Return value:	void
**
** IMPLICIT INPUTS:
**
**	NONE
**
** IMPLICIT OUTPUTS:
**
**	'output_file', 'latex_file'
**
** SIDE EFFECTS:
**
**	If the session logfile cannot be successfully opened, the function
**	exits with an error code.
**	As described in `show_menu', the order of options displayed in
**	that function is implied here.	If the order in `show_menu' were
**	to change then the `switch (selection)' case labels would have to
**	be adjusted accordingly. Selecting menu option 1 (edit) has no effect. 
*/

void SP_main()
{
	int i ;
	
	if ((input_file = fopen(in_filename, "r")) == NIL) 
		abort_run("Fatal: Unable to open input_file\n") ;
	if ((parameters_file = fopen(parameters_filename, "r")) == NIL)
		abort_run("Unable to open parameters file.\n");
	if ((output_file = fopen(out_filename, "w")) == NIL)
		abort_run("Unable to open write file.\n");
	if ((latex_file = fopen(latex_filename, "w")) == NIL) 
		abort_run("Fatal: Unable to open latex_file\n") ;
	if ((plot_file = fopen(plot_filename, "w")) == NIL) 
		abort_run("Fatal: Unable to open plot_file\n") ;

	fprintf(output_file, "%s%s%s%s%s%s%s%s",
		"SP70 version ", VERSION, ". Copyright (C) 2001 J G Wolff.\n",
		"This program is released to the public domain without\n",
		"any restrictions. You can redistribute it and/or modify\n",
		"it as you wish. Please retain this notice on all\n",
		"versions and acknowledge J G Wolff as the originator\n",
		"of the program in all publications.\n\n") ;

	fprintf(output_file, "%s%s%s%s%s",
		"RESULTS FROM ", PROGRAM, ", VERSION ", VERSION, "\n\n") ;
	
	time_now() ;
	
	fprintf(output_file, "%s%s%s",
		"INPUT FILE: ", in_filename, "\n\n") ;

	// Read parameters.

	read_parameters() ;
	
	half_hit_structure_rows = hit_structure_rows / 2 ;

	// Print out parameters.
	
	fprintf(output_file, "PARAMETERS:\n\n") ;
	
	fprintf(output_file, "%s%d%s", 
		"PROBABILITIES = ", probabilities, "\n\n") ;
	fprintf(output_file, "%s%d%s", 
		"SHOW_AL_STRUCTURE = ", show_al_structure, "\n\n") ;
	fprintf(output_file, "%s%d%s", 
		"SHOW_H_STRUCTURE = ", show_h_structure, "\n\n") ;
	fprintf(output_file, "%s%d%s", 
		"SHOW_UNSELECTED_ALIGNMENTS = ",
		show_unselected_alignments, "\n\n") ;
	fprintf(output_file, "%s%d%s", 
		"VERBOSE = ",
		verbose, "\n\n") ;
	fprintf(output_file, "%s%d%s", 
		"USE_ABBREVIATIONS = ",
		use_abbreviations, "\n\n") ;
	fprintf(output_file, "%s%d%s", 
		"SHOW_ALL_PARSING_ALIGNMENTS = ",
		show_all_parsing_alignments, "\n\n") ;
	fprintf(output_file, "%s%c%s", 
		"ALIGNMENT_FORMAT = ",
		alignment_format, "\n\n") ;
	fprintf(output_file, "%s%d%s", 
		"HIT_STRUCTURE_ROWS = ", hit_structure_rows, "\n\n") ;
	fprintf(output_file, "%s%d%s", 
		"WINDOW_SIZE = ", window_size, "\n\n") ;
	fprintf(output_file, "%s%d%s", 
		"FAIL_SCORE = ", fail_score, "\n\n") ;
	fprintf(output_file, "%s%1.2f%s", 
		"COST_FACTOR = ", cost_factor, "\n\n") ;
	fprintf(output_file, "%s%d%s", 
		"KEEP_ROWS = ", keep_rows, "\n\n") ;
	fprintf(output_file, "%s%d%s", 
		"DRIVING_KEEP_ROWS = ",
		driving_keep_rows, "\n\n") ;
	fprintf(output_file, "%s%d%s",
		"FULL_ALIGNMENT_LIMIT = ",
		full_alignment_limit, "\n\n") ;
	fprintf(output_file, "%s%d%s",
		"GRAMMAR_LIMIT = ",
		grammar_limit, "\n\n") ;
	fprintf(output_file, "%s%d%s",
		"EXTRACTION_LIMIT = ",
		extraction_limit, "\n\n") ;
	fprintf(output_file, "%s%s%s", 
		"FIGURE_ID = ",
		figure_ID, "\n\n") ;
	fprintf(output_file, "%s%d%s", 
		"MAX_NEW_GAP = ", 
		max_new_gap, "\n\n") ;
	fprintf(output_file, "%s%d%s", 
		"MAX_OLD_GAP = ", 
		max_old_gap, "\n\n") ;
	fprintf(output_file, "%s%d%s", 
		"MAX_DRIVING_GAP = ",
		max_driving_gap, "\n\n") ;
	fprintf(output_file, "%s%d%s", 
		"MAX_TARGET_GAP = ",
		max_target_gap, "\n\n") ;
	fprintf(output_file, "%s%d%s", 
		"MAX_UNSUPPORTED_CYCLES = ",
		max_unsupported_cycles, "\n\n") ;
	fprintf(output_file, "%s%d%s", 
		"MAX_ALIGNMENTS_IN_ONE_CYCLE = ",
		max_alignments_in_one_cycle, "\n\n") ;
	fprintf(output_file, "%s%d%s", 
		"ORIENTATION = ",
		orientation, "\n\n") ;
	fprintf(output_file, "%s%d%s", 
		"WRITE_AL_CHARS_LENGTH = ",
		write_al_chars_length, "\n\n") ;
	fprintf(output_file, "%s%d%s", 
		"WRITE_SECTION_CHARS_LENGTH = ",
		write_section_chars_length, "\n\n") ;
	fprintf(output_file, "%s%1.2f%s", 
		"MINIMUM_FONT_SIZE = ",
		minimum_font_height, "\n\n") ;
	fprintf(output_file, "%s%1.2f%s", 
		"MAXIMUM_FONT_SIZE = ",
		maximum_font_height, "\n\n") ;
	fprintf(output_file, "%s%d%s", 
		"MAXIMUM_LEARNING_COUNT = ",
		maximum_learning_count, "\n\n") ;
	
	/* Load SP structure */
	
	load() ;
	
	/* The first object in the corpus is taken to be 'New' */
	
	corpus->initialise() ;
	new_patterns = (group *)corpus->get_next_child() ;
	if (new_patterns == NIL) abort_run("main(): invalid New object") ;
	
	old_patterns = (group *)corpus->get_next_child() ;
	if (old_patterns == NIL) abort_run("main(): invalid Old object") ;

	// Internal constants not set in input file.
	
	fprintf(output_file, "%s%d%s", 
		"MAX_ALIGNMENT_LENGTH = ",
		max_alignment_length, "\n\n") ;
	
	fflush(output_file) ;
	
	if (fe_key_array == 0 && use_abbreviations != OFF) // fe_key_array
		// is used to detect whether a key to the abbreviations 
		// of symbols has been supplied or not.
	{
		if (use_abbreviations == LETTERS)
			make_letter_abbreviations() ;
		
		// if (use_abbreviations == DIGITS) 
		//	make_digit_abbreviations() ; Development of the
		// use of digit abbreviations has been halted.
	}
	
	write_key() ; // Writes out the key to abbreviations of symbols
	// used in the corpus.
		
	// Make global objects and arrays.
		
	hn_master = new hit_node ;
	hit_root = new hit_node ;
	current_alignments_from_parsing = new group ;
	full_alignments_from_parsing = new group ;
	global_alignments_from_parsing = new group ;
	keep_alignments = new group ;
	driving_patterns_and_alignments = new group ;
	created_alignments = new group ;
	created_code_patterns = new group ;
	symbols_in_new = new_patterns->symbol_set() ;
	set_of_grammars = new group ;
	set_of_cleaned_up_grammars = new group ;
	copies_of_grammar_patterns = new group ;
	naive_grammar = new grammar ;
	set_of_class_records = new group ;

	write_al_chars = new char*[MAX_ALIGNMENT_DEPTH] ;
	write_al_vacant_slots = new char*[MAX_ALIGNMENT_DEPTH] ;

	for (i = 0; i < MAX_ALIGNMENT_DEPTH; i++)
	{
		write_al_chars[i] = new char[write_al_chars_length] ;
		write_al_vacant_slots[i] = new char[write_al_chars_length] ;
	}

	repeat_cpas() ; // Iteratively create_new_patterns_and_sort() until
		// No more compression can be achieved.
		
	exit_routine(0) ;
} /* main */